import plotly.express as px
from dash import MATCH, Input, Output, ctx, dcc, html
from dash.exceptions import PreventUpdate

from xiplot.plots import APlot
from xiplot.plugin import ID_HOVERED
from xiplot.utils.auxiliary import decode_aux, merge_df_aux
from xiplot.utils.cluster import KMeans
from xiplot.utils.components import (
    ColumnDropdown,
    FlexRow,
    PdfButton,
    PlotData,
)
from xiplot.utils.dataframe import get_numeric_columns
from xiplot.utils.layouts import layout_wrapper
from xiplot.utils.regex import get_columns_by_regex


class Heatmap(APlot):
    @classmethod
    def register_callbacks(cls, app, df_from_store, df_to_store):
        PdfButton.register_callback(app, cls.name(), {"type": "heatmap"})

        @app.callback(
            Output({"type": "heatmap", "index": MATCH}, "figure"),
            Input({"type": "heatmap_cluster_amount", "index": MATCH}, "value"),
            Input(cls.get_id(MATCH, "columns_dropdown"), "value"),
            Input(ID_HOVERED, "data"),
            Input("data_frame_store", "data"),
            Input("auxiliary_store", "data"),
            Input("plotly-template", "data"),
            prevent_initial_call=False,
        )
        def tmp(n_clusters, features, hover, df, aux, template=None):
            # Try branch for testing
            try:
                if ctx.triggered_id == "data_frame_store":
                    raise PreventUpdate()
            except PreventUpdate:
                raise
            except Exception:
                pass

            return Heatmap.render(
                n_clusters,
                features,
                hover,
                df_from_store(df),
                decode_aux(aux),
                template,
            )

        ColumnDropdown.register_callback(
            app,
            cls.get_id(MATCH, "columns_dropdown"),
            df_from_store,
            numeric=True,
            regex_button_id=cls.get_id(MATCH, "regex_button"),
            regex_input_id=cls.get_id(MATCH, "regex_input"),
        )

        PlotData.register_callback(
            cls.name(),
            app,
            dict(
                clusters=Input(
                    {"type": "heatmap_cluster_amount", "index": MATCH}, "value"
                ),
                columns=Input(cls.get_id(MATCH, "columns_dropdown"), "value"),
            ),
        )

        return [tmp]

    @staticmethod
    def render(n_clusters, features, hover, df, aux, template=None):
        km = KMeans(n_clusters=n_clusters, random_state=42)
        df = merge_df_aux(df, aux)

        features = get_columns_by_regex(df.columns.to_list(), features)
        features = features if features else df.columns.to_list()
        features = get_numeric_columns(df, features)
        dff = df[features].dropna()
        km.fit(dff)

        fig = px.imshow(
            km.cluster_centers_,
            x=features,
            y=list(range(1, n_clusters + 1)),
            color_continuous_scale="RdBu",
            origin="lower",
            aspect="auto",
            template=template,
        )

        if hover is not None:
            # Due to dropna
            index = dff.index.searchsorted(hover)
            if index < len(dff.index) and dff.index[index] == hover:
                fig.add_hline(
                    km.labels_[index] + 1,
                    line=dict(color="rgba(0.5,0.5,0.5,0.5)", dash="dash"),
                )

        return fig

    @classmethod
    def create_layout(cls, index, df, columns, config=dict()):
        import jsonschema

        jsonschema.validate(
            instance=config,
            schema=dict(
                type="object", properties=dict(clusters=dict(type="integer"))
            ),
        )
        n_clusters = config.get("clusters", 5)
        num_columns = get_numeric_columns(df, columns)
        return [
            dcc.Graph(id={"type": "heatmap", "index": index}),
            FlexRow(
                layout_wrapper(
                    component=ColumnDropdown(
                        cls.get_id(index, "columns_dropdown"),
                        options=num_columns,
                        multi=True,
                        clearable=False,
                        value=config.get("columns", None),
                    ),
                    title="Features",
                    css_class="dash-dropdown",
                ),
                html.Button(
                    "Add features by regex",
                    id=cls.get_id(index, "regex_button"),
                    className="button",
                ),
                layout_wrapper(
                    component=dcc.Slider(
                        min=2,
                        max=10,
                        step=1,
                        value=n_clusters,
                        id={"type": "heatmap_cluster_amount", "index": index},
                    ),
                    title="Number of clusters",
                    css_class="dash-dropdown",
                ),
            ),
            dcc.Input(
                id=cls.get_id(index, "regex_input"), style={"display": "none"}
            ),
        ]



import re
from typing import List, Optional, Tuple


def dropdown_regex(
    options: List[str], selected: List[str], new_regex: Optional[str] = None
) -> Tuple[List[str], List[str], int]:
    """Dropdown options with regex.

    Args:
        options: List of dropdown options.
        selected: List of dropdown selection.
        new_regex: Regex to add to the selection. Defaults to None.

    Returns:
        new_options: New list of options (reduced by regexes).
        new_selected: New list of selection (including the new regex).
        hits: number of hits for the new regex or zero.
    """
    if selected is None:
        selected = []
    hits = False
    # Add optional new regex
    if new_regex:
        new_regex = new_regex + " (regex)"
        if new_regex not in selected:
            selected = selected + [new_regex]
            hits = True
    # Remove selected from options
    options = [o for o in options if o not in selected]
    old_len = len(options)
    # Remove regex matches from options
    is_regex = re.compile(r" \(regex(()|(: \d+))\)$")
    select_results = []
    for s in selected:
        reg = is_regex.search(s)
        if reg:
            s = s[: reg.start()]
            regex = re.compile(s)
            old_len = len(options)
            options = [o for o in options if not regex.search(o)]
            s = f"{s} (regex: {old_len - len(options)})"
        select_results.append(s)
    # Return new_options, new_selected, hits for the new regex
    return (
        select_results + options,
        select_results,
        old_len - len(options) if hits else 0,
    )


def get_columns_by_regex(columns, features):
    if features is None:
        return []

    new_features = []

    is_regex = re.compile(r" \(regex(()|(: \d+))\)$")
    for f in features:
        reg = is_regex.search(f)
        if reg:
            regex = re.compile(f[: reg.start()])
            for c in columns:
                if regex.search(c) and c not in new_features:
                    new_features.append(c)
        elif f not in new_features:
            new_features.append(f)
    return new_features


if __name__ == "__main__":
    print(
        dropdown_regex(
            [
                "PCA 1",
                "PCA 2",
                "mpg",
                "cylinders",
                "displacement",
                "horsepower",
                "weight",
                "acceleration",
                "model-year",
                "origin",
            ],
            ["mpg", "weight", "P.* (regex)"],
        )
    )



from dash import html


def layout_wrapper(component, id="", style=None, css_class=None, title=None):
    """
    Wraps a dash component to a html.Div element

    Parameters:

        component: dash core component
        id: id of the html.Div element
        style: style of the html.Div element
        css_class: className of the html.Div
        title: title of the dash core component

    Returns:

        layout: html.Div element
    """
    layout = html.Div(
        children=[html.Div(title), component],
        id=id,
        style=style,
        className=css_class,
    )
    return layout



from abc import ABC, abstractclassmethod
from typing import Any, Callable, Dict, List, Optional

import pandas as pd
from dash import Dash, dcc, html

from xiplot.utils import generate_id
from xiplot.utils.components import (
    DeleteButton,
    FlexRow,
    HelpButton,
    PdfButton,
    PlotData,
)


class APlot(ABC):
    """Abstract class that defines the API for implementing a new plot"""

    def __init__(self):
        raise TypeError("APlot should not be constructed")

    @classmethod
    def name(cls) -> str:
        """The name that is shown in the UI when selecting plots."""
        return cls.__name__

    @classmethod
    def help(cls) -> Optional[str]:
        """Tooltip that describes the plot and how to use it."""
        # Recommended format: "Short description.\n\nLong description."
        return None

    @classmethod
    def get_id(
        cls, index: Any, subtype: Optional[str] = None
    ) -> Dict[str, Any]:
        """Generate id:s for the plot.

        Args:
            index: Index of the plot.
            subtype: Differentiatie parts of the plot. Defaults to None.

        Returns:
            Dash id.
        """
        return generate_id(cls, index, subtype)

    @abstractclassmethod
    def register_callbacks(
        cls,
        app: Dash,
        df_from_store: Callable[[Any], pd.DataFrame],
        df_to_store: Callable[[pd.DataFrame], Any],
    ):
        """Override this to register Dash callbacks for your plot.

        If a PdfButton is added to the layout, remember to do:
        `PdfButton.register_callback(app, cls.name() cls.get_id(None))`
        If a PlotData is added to the layout, remember to do:
        `PlotData.register_callback(app, cls.name(), ...)`

        Args:
            app: The xiplot app (a subclass of `dash.Dash`).
            df_from_store: Functions that transforms `dcc.Store` data into a
                dataframe.
            df_to_store: Function that transform a dataframe into `dcc.Store`
                data.
        """
        pass

    @abstractclassmethod
    def create_new_layout(
        cls,
        index: Any,
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
        config: Dict[str, Any] = dict(),
    ) -> html.Div:
        """Overide this method to create a layout for your plot.
        Alternatively you can overload `APlot.create_layout` for a simplified
        setup.

        Args:
            index: The index of the plot.
            df: Dataframe.
            columns: Columns from the dataframe to use in the plot.
            config: Plot configuration (used when restoring saved plots).
                Defaults to dict().

        Returns:
            A html element presenting the plot.
        """
        children = cls.create_layout(index, df, columns, config)
        top_bar = [DeleteButton(index), html.Div(className="stretch")]
        if any(isinstance(e, dcc.Graph) for e in children):
            top_bar.append(PdfButton(index, cls.name()))
        if cls.help():
            top_bar.append(HelpButton(cls.help()))
        children.insert(0, FlexRow(*top_bar))
        children.append(PlotData(index, cls.name()))
        return html.Div(
            children, id=cls.get_id(index, "panel"), className="plots"
        )

    @abstractclassmethod
    def create_layout(
        cls,
        index: Any,
        df: pd.DataFrame,
        columns: Any,
        config: Dict[str, Any] = dict(),
    ) -> List[Any]:
        """If `APlot.create_new_layout` is not overriden, then this method can
        be overriden to provide a "standard" plot.

        A "standard" plot has the following features:
            - The children (given by this function) are wrapped in a Div with
              `className="plots"`.
            - A `DeleteButton` is added, so that the plot can be removed.
            - A `PdfButton` is added if any of the children is a `dcc.Graph`.
            - A `PlotData` is added, so that the plot can be saved.

        Args:
            index: The index of the plot
            df: Dataframe.
            columns: Columns from the dataframe to use in the plot.
            config: Plot configuration (used when restoring saved plots).
                Defaults to dict().

        Returns:
            A list of (child) html components.
        """
        return ["Not implemented"]



def cli():
    import argparse
    import os

    from xiplot.setup import setup_xiplot_dash_app

    parser = argparse.ArgumentParser(
        prog="xiplot",
        description="Xiplot:   A Dash app for interactively visualising data",
    )
    parser.add_argument(
        "PATH",
        nargs="?",
        default="data",
        help="The path to a directory containing data files",
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Enable debug mode"
    )
    parser.add_argument(
        "-p", "--port", help="Port used to serve the application"
    )
    parser.add_argument("--host", help="Host IP used to serve the application")
    parser.add_argument(
        "-c",
        "--cache",
        action="store_true",
        help=(
            "Cache datasets on the server in order to reduce the amount of"
            " data transferred. Might not be suitable for servers with"
            " multiple users"
        ),
    )
    parser.add_argument(
        "--plugin", help="The path to a directory containing plugin .whl files"
    )
    args = parser.parse_args()
    path = args.PATH

    if not os.path.isdir(path):
        print(f'Directory "{path}" was not found')

    kwargs = {}
    if args.debug:
        kwargs["debug"] = True
    if args.host:
        kwargs["host"] = args.host
    if args.port:
        kwargs["port"] = args.port
    if args.plugin:
        plugin_dir = args.plugin
    else:
        plugin_dir = "plugins"

    unsafe_local_server = True if args.cache else False

    app = setup_xiplot_dash_app(
        unsafe_local_server=unsafe_local_server,
        data_dir=path,
        plugin_dir=plugin_dir,
    )
    app.run(**kwargs)



import sys
import uuid

import dash
import dash_mantine_components as dmc
from dash import Input, Output, State, dcc, html

from xiplot.tabs import Tab
from xiplot.utils.auxiliary import decode_aux, encode_aux, merge_df_aux
from xiplot.utils.components import ColumnDropdown, FlexRow
from xiplot.utils.layouts import layout_wrapper
from xiplot.utils.regex import get_columns_by_regex


def get_pca_columns(df, features):
    from sklearn.decomposition import PCA
    from sklearn.impute import SimpleImputer
    from sklearn.preprocessing import StandardScaler

    x = df[features]
    x = StandardScaler().fit_transform(x)

    mean_imputer = SimpleImputer(strategy="mean")
    x = mean_imputer.fit_transform(x)

    pca = PCA(n_components=2)
    pca.fit(x)
    return pca.transform(x)


class Embedding(Tab):
    def register_callbacks(app, df_from_store, df_to_store):
        @app.callback(
            Output("auxiliary_store", "data"),
            Output("embedding-tab-main-notify-container", "children"),
            Output("embedding-tab-compute-done", "children"),
            Input("embedding-button", "value"),
            State("data_frame_store", "data"),
            State("auxiliary_store", "data"),
            State("embedding_feature", "value"),
            State("embedding_type", "value"),
        )
        def compute_embedding(process_id, df, aux, features, embedding_type):
            if df is None:
                return (
                    dash.no_update,
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message="You have not yet loaded any data file.",
                        action="show",
                        autoClose=10000,
                    ),
                    dash.no_update,
                )

            df = df_from_store(df)
            aux = decode_aux(aux)
            columns = ColumnDropdown.get_columns(df, aux, numeric=True)
            features = get_columns_by_regex(columns, features)

            notifications = []
            if not Embedding.validate_embedding_params(
                embedding_type, features, notifications, process_id
            ):
                return dash.no_update, notifications, process_id

            try:
                pca_cols = get_pca_columns(merge_df_aux(df, aux), features)
                aux["Xiplot_PCA_1"] = pca_cols[:, 0]
                aux["Xiplot_PCA_2"] = pca_cols[:, 1]

                notifications.append(
                    dmc.Notification(
                        id=process_id or str(uuid.uuid4()),
                        color="green",
                        title="Success",
                        message="The data was embedded successfully!",
                        action="update" if process_id else "show",
                        autoClose=5000,
                        disallowClose=False,
                    )
                )

            except ImportError as err:
                raise err

            except Exception as err:
                notifications.append(
                    dmc.Notification(
                        id=process_id or str(uuid.uuid4()),
                        color="red",
                        title="Error",
                        message=(
                            "The embedding failed with an internal error."
                            f" Please report the following bug: {err}"
                        ),
                        action="update" if process_id else "show",
                        autoClose=False,
                    )
                )

                return (dash.no_update, notifications, process_id)

            return encode_aux(aux), notifications, process_id

        @app.callback(
            Output("embedding-button", "value"),
            Output("embedding-tab-compute-notify-container", "children"),
            Input("embedding-button", "n_clicks"),
            State("embedding_type", "value"),
            State("embedding_feature", "value"),
            State("embedding-tab-compute-done", "children"),
            State("embedding-button", "value"),
        )
        def compute_embeddings(
            n_clicks, embedding_type, features, done, doing
        ):
            if done != doing:
                return dash.no_update, dmc.Notification(
                    id=str(uuid.uuid4()),
                    color="yellow",
                    title="Warning",
                    message=(
                        "The k-means clustering process has not yet finished."
                    ),
                    action="show",
                    autoClose=10000,
                )

            process_id = str(uuid.uuid4())

            message = "The embedding process has started."

            if "sklearn" not in sys.modules:
                message += " [Loading scikit-learn]"

            return process_id, dmc.Notification(
                id=process_id,
                color="blue",
                title="Processing",
                message=message,
                action="show",
                loading=True,
                autoClose=False,
                disallowClose=True,
            )

        ColumnDropdown.register_callback(
            app,
            "embedding_feature",
            df_from_store,
            numeric=True,
            regex_button_id="embedding-regex-button",
            regex_input_id="embedding-regex-input",
        )

    @staticmethod
    def validate_embedding_params(
        embedding_type, features, notifications=None, process_id=None
    ):
        if features is None:
            if notifications is not None and process_id is not None:
                notifications.append(
                    dmc.Notification(
                        id=process_id,
                        action="hide",
                    )
                )

            if features is None and notifications is not None:
                notifications.append(
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message=(
                            "You have not selected any features to embed by."
                        ),
                        action="show",
                        autoClose=10000,
                    )
                )

            if embedding_type is None and notifications is not None:
                notifications.append(
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message="You have not selected embedding type.",
                        action="show",
                        autoClose=10000,
                    )
                )

            return False

        if len(features) < 2 and embedding_type == "PCA":
            if notifications is not None and process_id is not None:
                notifications.append(
                    dmc.Notification(
                        id=process_id,
                        action="hide",
                    )
                )

            if notifications is not None:
                notifications.append(
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message="You must select at least two features",
                        action="show",
                        autoClose=10000,
                    )
                )

            return False

        return True

    @staticmethod
    def create_layout():
        return FlexRow(
            layout_wrapper(
                dcc.Dropdown(
                    id="embedding_type",
                    options=["PCA"],
                    value="PCA",
                    clearable=False,
                ),
                css_class="dash-dropdown",
                title="Embedding type",
            ),
            layout_wrapper(
                component=ColumnDropdown(id="embedding_feature", multi=True),
                css_class="dash-dropdown",
                title="Features",
            ),
            dcc.Input(id="embedding-regex-input", style={"display": "none"}),
            html.Button(
                "Add features by regex",
                id="embedding-regex-button",
                className="button",
            ),
            html.Button(
                "Compute the embedding",
                id="embedding-button",
                className="button",
            ),
            id="control-embedding-container",
        )

    @staticmethod
    def create_layout_globals():
        return html.Div(
            [
                html.Div(
                    id="embedding-tab-main-notify-container",
                    style={"display": "none"},
                ),
                html.Div(
                    id="embedding-tab-compute-notify-container",
                    style={"display": "none"},
                ),
                html.Div(
                    id="embedding-tab-compute-done", style={"display": "none"}
                ),
            ],
            style={"display": "none"},
        )



import dash
import pandas as pd
from dash import MATCH, Input, Output, State, dcc, html

from xiplot.plots import APlot
from xiplot.plugin import ID_CLICKED, ID_DATAFRAME, ID_HOVERED
from xiplot.utils.components import FlexRow, InputText, PlotData
from xiplot.utils.layouts import layout_wrapper


class Smiles(APlot):
    @classmethod
    def name(cls):
        return "SMILES (molecules)"

    @classmethod
    def help(cls):
        return (
            "Render a molecule from a SMILES string\n\nOne column of the"
            " dataset must contain a molecule represented as a SMILES string"
            " (Simplified molecular-input line-entry system)."
        )

    @classmethod
    def register_callbacks(cls, app, df_from_store, df_to_store):
        app.clientside_callback(
            """
async function svgFromSMILES(smiles) {
    if (!window.RDKit) {
        window.RDKit = await window.initRDKitModule();
    }
    const INVALID_SVG = `
<?xml version='1.0' encoding='iso-8859-1'?>
<svg version='1.1' baseProfile='full'
    xmlns='http://www.w3.org/2000/svg'
    xmlns:rdkit='http://www.rdkit.org/xml'
    xmlns:xlink='http://www.w3.org/1999/xlink'
    xml:space='preserve'
    width='250px' height='200px' viewBox='0 0 250 200'
>
    <line x2='200' y2='175' x1='50' y1='25' stroke='#c0392b' stroke-width='5'/>
    <line x2='50' y2='175' x1='200' y1='25' stroke='#c0392b' stroke-width='5'/>
</svg>
    `;
    const mol = window.RDKit.get_mol(smiles);
    const svg = smiles && mol
        .get_svg()
        .replace(/<rect[^>]*>\\s*<\\/rect>/, "")
        .split(/\\s+/)
        .join(" ");

    return "data:image/svg+xml;base64," + btoa(svg || INVALID_SVG);
}
            """,
            Output(cls.get_id(MATCH, "display"), "src"),
            Input(cls.get_id(MATCH, "string"), "value"),
            prevent_initial_call=False,
        )

        @app.callback(
            Output(cls.get_id(MATCH, "string"), "value"),
            Input(ID_CLICKED, "data"),
            Input(ID_HOVERED, "data"),
            State(cls.get_id(MATCH, "mode"), "value"),
            State(cls.get_id(MATCH, "col"), "value"),
            State(cls.get_id(MATCH, "string"), "value"),
            State(ID_DATAFRAME, "data"),
            prevent_initial_call=True,
        )
        def update_smiles(rowc, rowh, mode, col, old, df):
            if col is None or col == "":
                return dash.no_update
            if mode == "Click":
                row = rowc
            elif mode == "Hover":
                row = rowh
            else:
                raise Exception("Unknown SMILES mode: " + mode)
            if row is None:
                return dash.no_update
            df = df_from_store(df)
            try:
                new = df[col][row]
                if old != new:
                    return new
                return dash.no_update
            except Exception:
                return dash.no_update

        PlotData.register_callback(
            cls.name(),
            app,
            dict(
                mode=Input(cls.get_id(MATCH, "mode"), "value"),
                smiles=Input(cls.get_id(MATCH, "string"), "value"),
                column=Input(cls.get_id(MATCH, "col"), "value"),
            ),
        )

        return update_smiles

    @classmethod
    def create_layout(cls, index, df: pd.DataFrame, columns, config=dict()):
        import jsonschema

        jsonschema.validate(
            instance=config,
            schema=dict(
                type="object",
                properties=dict(
                    mode=dict(enum=["Hover", "Click"]),
                    smiles=dict(type="string"),
                    column=dict(type="string"),
                ),
            ),
        )

        cols = [
            c
            for c in df.select_dtypes([object, "category"])
            if isinstance(df[c][0], str)
        ]
        column = next((c for c in cols if "smiles" in c.lower()), "")

        render_mode = config.get("mode", "Hover")
        smiles_input = config.get("smiles", "")
        column = config.get("column", column)

        return [
            html.Img(
                id=cls.get_id(index, "display"),
                className="smiles-img",
            ),
            html.Br(),
            FlexRow(
                layout_wrapper(
                    dcc.Dropdown(
                        id=cls.get_id(index, "col"), value=column, options=cols
                    ),
                    title="SMILES column",
                    css_class="dash-dropdown",
                ),
                layout_wrapper(
                    dcc.Dropdown(
                        id=cls.get_id(index, "mode"),
                        value=render_mode,
                        options=["Hover", "Click"],
                        clearable=False,
                        searchable=False,
                    ),
                    title="Selection mode",
                    css_class="dash-dropdown",
                ),
                layout_wrapper(
                    InputText(
                        id=cls.get_id(index, "string"),
                        value=smiles_input,
                        debounce=True,
                        placeholder="SMILES string, e.g., 'O.O[Fe]=O'",
                    ),
                    css_class="dash-input",
                    title="SMILES string",
                ),
            ),
        ]



import warnings

from xiplot.tabs.plugins import get_plugins_cached
from xiplot.utils.dataframe import read_functions, write_functions


def test_plugin():
    try:
        import xiplot_test_plugin  # noqa: F401
    except ModuleNotFoundError:
        warnings.warn(
            "The test plugin is not installed, skipping plugin tests"
        )
        return

    # Remember to update this test if anything is added to the test_plugin.
    try:
        from xiplot_test_plugin import (
            Plot,
            create_global,
            plugin_load,
            plugin_write,
            register_callbacks,
        )
    except ImportError:
        warnings.warn(
            "The test plugin needs to be updated: "
            "`pip install --upgrade ./test_plugin`"
        )
        raise

    # Writing
    assert any(
        plugin == plugin_load for (_, _, plugin) in get_plugins_cached("read")
    )
    assert any(ext == plugin_load()[1] for _, ext in read_functions())

    # Reading
    assert any(
        plugin == plugin_write
        for (_, _, plugin) in get_plugins_cached("write")
    )
    assert any(ext == plugin_write()[1] for _, ext, _ in write_functions())

    # Plotting
    assert any(plugin == Plot for (_, _, plugin) in get_plugins_cached("plot"))
    assert any(
        plot.name() == Plot.name()
        for (_, _, plot) in get_plugins_cached("plot")
    )

    # Globals
    assert any(
        plugin == create_global
        for (_, _, plugin) in get_plugins_cached("global")
    )
    text = create_global().children[1].id["type"]
    assert any(
        text in str(g().children) for (_, _, g) in get_plugins_cached("global")
    )

    # Callbacks
    assert any(
        plugin == register_callbacks
        for (_, _, plugin) in get_plugins_cached("callback")
    )
    assert any(
        cb.__module__ == "xiplot_test_plugin"
        for (_, _, cb) in get_plugins_cached("callback")
    )



import pandas as pd

from xiplot.utils.auxiliary import decode_aux, encode_aux, toggle_selected
from xiplot.utils.components import ColumnDropdown
from xiplot.utils.regex import dropdown_regex, get_columns_by_regex


def test_add_matching_values():
    d = {"col1": [1, 2], "col2": [3.0, 4.0], "col3": ["a", "b"]}
    df = pd.DataFrame(data=d)
    assert ["col0", "col1", "col2", "col3"] == ColumnDropdown.get_columns(
        df, pd.DataFrame(), ["col0"]
    )
    assert ["col1", "col2"] == ColumnDropdown.get_columns(
        df, pd.DataFrame(), numeric=True
    )
    assert ["col1", "col3"] == ColumnDropdown.get_columns(
        df, pd.DataFrame(), category=True
    )


def test_regex():
    options, selected, hits = dropdown_regex(
        ["asd", "bsd", "csd", "dsd"], ["asd", "bs (regex)"], "cs"
    )
    assert options == ["asd", "bs (regex: 1)", "cs (regex: 1)", "dsd"]
    assert options[:-1] == selected
    assert hits == 1

    cols = get_columns_by_regex(
        ["asd", "bsd", "csd", "dsd"], ["asd", "(b|c)sd (regex)"]
    )
    assert cols == ["asd", "bsd", "csd"]


def test_aux():
    df = pd.DataFrame(index=range(4))
    assert df.equals(decode_aux(encode_aux(df)))
    df = pd.DataFrame(
        {"a": [1, 2, 3], "b": [1.0, 2.3, 3.2], "c": ["a", "b", "a"]}
    )
    assert df.equals(decode_aux(encode_aux(df)))
    df = toggle_selected(df, [])
    assert df.equals(toggle_selected(toggle_selected(df, 1), [1]))



from io import BytesIO

import pandas as pd

try:
    from xiplot_filetypes import (
        read_feather,
        read_parquet,
        write_feather,
        write_parquet,
    )
except ImportError:
    import sys
    from pathlib import Path

    sys.path.insert(
        0, str(Path(__file__).parent.parent / "plugin_xiplot_filetypes")
    )
    from xiplot_filetypes import (
        read_feather,
        read_parquet,
        write_feather,
        write_parquet,
    )


def test_feather():
    df = pd.DataFrame({"a": [1, 2, 3], "b": ["a", "b", "c"]})
    io = BytesIO()
    write_feather()[0](df, io)
    df2 = read_feather()[0](io)
    assert df.equals(df2)


def test_parquet():
    df = pd.DataFrame({"a": [1, 2, 3], "b": ["a", "b", "c"]})
    io = BytesIO()
    write_parquet()[0](df, io)
    df2 = read_parquet()[0](io)
    assert df.equals(df2)



import time

import dash
import pandas as pd
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from tests.util_test import render_plot, start_server
from xiplot.plots.table import Table
from xiplot.utils.auxiliary import SELECTED_COLUMN_NAME, get_selected

(
    update_table_data,
    update_selected_rows_store,
    update_lastly_activated_cell,
    update_table_columns,
) = Table.register_callbacks(dash.Dash(__name__), lambda x: x, lambda x: x)


def test_teta001_render_table(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Table")

    plot = driver.find_element(
        By.XPATH,
        "//table[@class='cell-table']",
    )

    assert "PCA 1" in plot.get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_teta002_select_columns(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Table")

    column_dropdown = driver.find_element(
        By.XPATH,
        "//div[@class='plots']/div[2]/div",
    )
    column_dropdown.click()

    column_dropdown_input = driver.find_element(
        By.XPATH,
        "//div[@class='plots']/div[2]/div/div[1]/div[1]/div[1]/div[2]/input",
    )
    column_dropdown_input.send_keys("PCA 2", Keys.RETURN)
    time.sleep(0.5)

    column_dropdown_button = driver.find_element(
        By.XPATH,
        "//button[text()='Update table']",
    )
    column_dropdown_button.click()
    time.sleep(0.5)

    plot = driver.find_element(
        By.XPATH,
        "//table[@class='cell-table']",
    )

    assert "PCA 1" not in plot.get_attribute("outerHTML")
    assert "PCA 2" in plot.get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_teta003_toggle_columns(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Table")

    toggle_columns = driver.find_element(
        By.XPATH,
        "//button[text()='Toggle Columns']",
    )
    toggle_columns.click()

    toggle_columns_first_checkbox = driver.find_element(
        By.XPATH, "//div[@class='show-hide-menu-item']/input"
    )
    toggle_columns_first_checkbox.click()

    plot = driver.find_element(
        By.XPATH,
        "//table[@class='cell-table']",
    )

    assert "PCA 1" not in plot.get_attribute("outerHTML")
    assert "PCA 2" in plot.get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_create_table():
    df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    aux = pd.DataFrame({SELECTED_COLUMN_NAME: [True, True]})
    output = update_table_data(aux, [df], [[]])
    table_df = output[0][0]
    sort_by = output[1][0]

    assert table_df[0] == {"col1": 1, "col2": 3, "Selection": True}
    assert table_df[1] == {"col1": 2, "col2": 4, "Selection": True}
    assert sort_by == []


def test_update_selected_rows_store():
    aux = pd.DataFrame({SELECTED_COLUMN_NAME: [False, False]})
    output = update_selected_rows_store([[1]], aux)
    assert all(get_selected(output) == [False, True])


def test_update_table_checkbox():
    df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    aux = pd.DataFrame({SELECTED_COLUMN_NAME: [False, True]})
    output = update_table_data(aux, [df], [[]])
    selected_rows = output[2]

    assert selected_rows == [[1]]


def test_update_lastly_activated_cell():
    d = {"col1": [1, 2], "col2": [3, 4]}
    df = pd.DataFrame(data=d)
    output = update_lastly_activated_cell(
        [{"row": 1, "column_id": "col1"}], [[0, 1]], df
    )
    lastly_activated_cell_row = output["cell_store"]
    updated_active_cell = output["active_cell"]

    assert lastly_activated_cell_row == 1
    assert updated_active_cell == [None]



import dash
import plotly.express as px
from dash import ALL, MATCH, Input, Output, State, dcc
from dash.exceptions import PreventUpdate

from xiplot.plots import APlot
from xiplot.plugin import (
    ID_AUXILIARY,
    ID_CLICKED,
    ID_HOVERED,
    placeholder_figure,
)
from xiplot.utils.auxiliary import (
    CLUSTER_COLUMN_NAME,
    SELECTED_COLUMN_NAME,
    decode_aux,
    merge_df_aux,
    toggle_selected,
)
from xiplot.utils.cluster import cluster_colours
from xiplot.utils.components import (
    ColumnDropdown,
    FlexRow,
    PdfButton,
    PlotData,
)
from xiplot.utils.dataframe import get_default_column
from xiplot.utils.layouts import layout_wrapper


class Boxplot(APlot):
    @classmethod
    def register_callbacks(cls, app, df_from_store, df_to_store):
        PdfButton.register_callback(app, cls.name(), cls.get_id(MATCH))

        @app.callback(
            Output(cls.get_id(MATCH), "figure"),
            Input(cls.get_id(MATCH, "x_axis"), "value"),
            Input(cls.get_id(MATCH, "y_axis"), "value"),
            Input(cls.get_id(MATCH, "plot"), "value"),
            Input(cls.get_id(MATCH, "color"), "value"),
            Input(ID_HOVERED, "data"),
            Input("data_frame_store", "data"),
            Input("auxiliary_store", "data"),
            Input("plotly-template", "data"),
            prevent_initial_call=False,
        )
        def render(
            x_axis,
            y_axis,
            plot,
            color,
            hover,
            df,
            aux,
            template=None,
        ):
            return cls.render(
                df_from_store(df),
                decode_aux(aux),
                x_axis,
                y_axis,
                plot,
                color,
                hover,
                template,
            )

        def get_row(hover):
            try:
                for p in hover:
                    if p is not None:
                        return p["points"][0]["customdata"][0]
            except (KeyError, TypeError):
                raise PreventUpdate()
            raise PreventUpdate()

        @app.callback(
            Output(ID_HOVERED, "data"),
            Output(cls.get_id(ALL), "hoverData"),
            Input(cls.get_id(ALL), "hoverData"),
        )
        def handle_hover_events(hover):
            return get_row(hover), [None] * len(hover)

        @app.callback(
            Output(ID_AUXILIARY, "data"),
            Output(ID_CLICKED, "data"),
            Output(cls.get_id(ALL), "clickData"),
            Input(cls.get_id(ALL), "clickData"),
            State(ID_AUXILIARY, "data"),
        )
        def handle_click_events(click, aux):
            row = get_row(click)
            if aux is None:
                return dash.no_update, row
            return toggle_selected(aux, row), row, [None] * len(click)

        PlotData.register_callback(
            cls.name(),
            app,
            {
                "x_axis": Input(cls.get_id(MATCH, "x_axis"), "value"),
                "y_axis": Input(cls.get_id(MATCH, "y_axis"), "value"),
                "color": Input(cls.get_id(MATCH, "color"), "value"),
                "plot": Input(cls.get_id(MATCH, "plot"), "value"),
            },
        )

        ColumnDropdown.register_callback(
            app, cls.get_id(ALL, "x_axis"), df_from_store, category=True
        )
        ColumnDropdown.register_callback(
            app, cls.get_id(ALL, "y_axis"), df_from_store, numeric=True
        )
        ColumnDropdown.register_callback(
            app, cls.get_id(ALL, "color"), df_from_store, category=True
        )

        return render

    @staticmethod
    def render(
        df,
        aux,
        x_axis,
        y_axis,
        plot="Box plot",
        color=None,
        hover=None,
        template=None,
    ):
        if y_axis is None:
            return placeholder_figure("Please select y axis")
        df = merge_df_aux(df, aux)
        df["__Xiplot_index__"] = range(df.shape[0])
        if x_axis not in df.columns:
            return placeholder_figure("Please select x axis")
        if color not in df.columns:
            color = None
        if plot == "Box plot":
            fig = px.box(
                df,
                x=x_axis,
                y=y_axis,
                color=color,
                notched=True,
                custom_data="__Xiplot_index__",
                color_discrete_map=cluster_colours(),
                template=template,
            )
            fig.update_traces(dict(boxmean=True))
        elif plot == "Violin plot":
            fig = px.violin(
                df,
                x=x_axis,
                y=y_axis,
                color=color,
                custom_data="__Xiplot_index__",
                color_discrete_map=cluster_colours(),
                template=template,
            )
        elif plot == "Strip chart":
            fig = px.strip(
                df,
                x=x_axis,
                y=y_axis,
                color=color,
                color_discrete_map=cluster_colours(),
                custom_data="__Xiplot_index__",
                template=template,
            )
        else:
            return placeholder_figure("Unsupported plot type")

        if hover is not None:
            fig.add_hline(
                df[y_axis][hover],
                line=dict(color="rgba(0.5,0.5,0.5,0.5)", dash="dash"),
                layer="below",
            )
        if SELECTED_COLUMN_NAME in aux:
            color = "#DDD" if template and "dark" in template else "#333"
            for x in df[y_axis][aux[SELECTED_COLUMN_NAME]]:
                fig.add_hline(
                    x, line=dict(color=color, width=0.5), layer="below"
                )
        return fig

    @classmethod
    def create_layout(cls, index, df, columns=None, config=dict()):
        num_columns = ColumnDropdown.get_columns(df, numeric=True)
        cat_columns = ColumnDropdown.get_columns(df, category=True)

        x_axis = config.get("x_axis", get_default_column(cat_columns, "x"))
        y_axis = config.get("y_axis", get_default_column(num_columns, "y"))
        color = config.get("color", CLUSTER_COLUMN_NAME)
        plot = config.get("plot", "Box plot")

        return [
            dcc.Graph(id=cls.get_id(index)),
            FlexRow(
                layout_wrapper(
                    component=ColumnDropdown(
                        cls.get_id(index, "x_axis"),
                        options=num_columns,
                        value=x_axis,
                        clearable=True,
                    ),
                    css_class="dash-dropdown",
                    title="X",
                ),
                layout_wrapper(
                    component=ColumnDropdown(
                        cls.get_id(index, "y_axis"),
                        options=num_columns,
                        value=y_axis,
                        clearable=False,
                    ),
                    css_class="dash-dropdown",
                    title="Y",
                ),
                layout_wrapper(
                    component=ColumnDropdown(
                        cls.get_id(index, "color"),
                        options=columns,
                        value=color,
                        clearable=True,
                    ),
                    css_class="dash-dropdown",
                    title="Groups",
                ),
                layout_wrapper(
                    component=dcc.Dropdown(
                        id=cls.get_id(index, "plot"),
                        options=["Box plot", "Violin plot", "Strip chart"],
                        value=plot,
                        clearable=False,
                    ),
                    css_class="dash-dropdown",
                    title="Plot type",
                ),
            ),
        ]



from io import BytesIO

import numpy as np
import pandas as pd

from xiplot.utils.dataframe import read_only_dataframe, write_functions


def test_read_write():
    orig = pd.DataFrame(
        dict(
            x=np.arange(5),
            y=[f"a{i}" for i in range(5)],
            b=np.arange(5) > 2,
            c=np.arange(1, 6),
        )
    )

    def write(fn, df=orig):
        bytes = BytesIO()
        fn(orig, bytes)
        return BytesIO(bytes.getvalue())

    for fn, ext, mime in write_functions():
        bytes1 = write(fn)
        try:
            read1 = read_only_dataframe(bytes1, "a" + ext)
        except Exception:
            print("Error when reading:", ext)
            raise
        if "example" not in mime:
            assert orig.equals(read1), "Could not handle: " + ext
        bytes2 = write(fn, read1)
        read2 = read_only_dataframe(bytes2, "b" + ext)
        assert read1.equals(read2), "Is not consistent: " + ext



import time

import dash
import pandas as pd
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from tests.util_test import render_plot, start_server
from xiplot.plots.scatterplot import Scatterplot
from xiplot.utils.auxiliary import get_clusters, get_selected

(
    tmp,
    handle_click_events,
    handle_hover_events,
    handle_cluster_drawing,
) = Scatterplot.register_callbacks(
    dash.Dash(__name__), lambda x: x, lambda x: x
)


def test_tesc001_render_scatterplot(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Scatterplot")

    plot = driver.find_element(By.CLASS_NAME, "dash-graph")

    assert "scatterplot" in plot.get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_tesc002_change_axis_value(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Scatterplot")

    driver.find_element(By.CLASS_NAME, "dd-double-left").click()

    x = driver.find_element(
        By.XPATH,
        (
            "//div[@class='dd-double-left']"
            "/div[2]/div[1]/div[1]/div[1]/div[2]/input"
        ),
    )

    x.send_keys("mpg")
    x.send_keys(Keys.RETURN)
    time.sleep(0.5)

    assert "mpg" in driver.find_element(By.CLASS_NAME, "xtitle").text
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_tesc003_target_setting(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Scatterplot")

    color = driver.find_element(
        By.XPATH,
        (
            "//div[@class='plots']/div[3]/div[2]/"
            "div[1]/div[1]/div[1]/div[2]/input"
        ),
    )

    color.send_keys("PCA 1")
    color.send_keys(Keys.RETURN)
    time.sleep(0.5)

    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_tesc004_jitter_setting(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Scatterplot")

    jitter_slider = driver.find_element(By.CLASS_NAME, "rc-slider-step")
    jitter_slider.click()
    time.sleep(0.5)

    jitter_value = driver.find_element(By.CLASS_NAME, "rc-slider-handle")

    assert "0.5" in jitter_value.get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"
    driver.close()


def test_create_scatterplot():
    fig = tmp(
        "col1",
        "col2",
        "Clusters",
        None,
        0,
        pd.DataFrame({"col1": [1, 2], "col2": [3, 4]}),
        pd.DataFrame(index=range(2)),
        None,
    )
    assert str(type(fig)) == "<class 'plotly.graph_objs._figure.Figure'>"


def test_handle_click_events():
    click = [{"points": [{"customdata": [{"index": 0}]}]}]
    output = handle_click_events(click, pd.DataFrame(index=range(2)))

    aux = output["aux"]
    clicked_row = output["click_store"]
    clicked_update = output["scatter"]

    assert all(get_selected(aux) == [True, False])
    assert clicked_row == 0
    assert clicked_update == [None]


def test_handle_hover_events():
    hover = [{"points": [{"customdata": [{"index": 1}]}]}]
    output = handle_hover_events(hover)

    hovered_row = output["hover_store"]
    hover_update = output["scatter"]

    assert hovered_row == 1
    assert hover_update == [None]


def test_handle_cluster_drawing():
    selected_points = [{"points": [{"customdata": [{"index": 1}]}]}]
    aux = handle_cluster_drawing(
        selected_points, pd.DataFrame(index=range(2)), "c1", False
    )
    assert all(get_clusters(aux) == ["c2", "c1"])



import multiprocessing
import sys


def check_loaded(queue, modules):
    import xiplot  # noqa: F401
    import xiplot.setup  # noqa: F401

    queue.put([mod in sys.modules for mod in modules])


def test_ensure_lazyload():
    # This tests that the following packages are not imported when xiplot is
    # imported, so that they can be lazily loaded in the WASM version.
    # This test starts a new Python (multi-)process using "spawn" to check the
    # imports in a clean environment.
    modules = ["sklearn", "jsonschema", "plotly.figure_factory"]
    ctx = multiprocessing.get_context("spawn")
    q = ctx.Queue()
    p = ctx.Process(target=check_loaded, args=(q, modules))
    p.start()
    for mod, loaded in zip(modules, q.get()):
        assert not loaded, f"{mod} should be lazily loaded"
    p.join()



import base64
import uuid
from collections import Counter
from io import BytesIO
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

import pandas as pd
import plotly as po
from dash import (
    ALL,
    MATCH,
    Dash,
    Input,
    Output,
    State,
    ctx,
    dcc,
    html,
    no_update,
)

from xiplot.utils import generate_id
from xiplot.utils.auxiliary import decode_aux, get_clusters
from xiplot.utils.cluster import cluster_colours
from xiplot.utils.regex import dropdown_regex


class FlexRow(html.Div):
    """
    Create a Div with a "display: flex" style.
    This makes the children appear on one line (with wrapping if necessary).
    Any child with a `className="stretch"` will fill all the empty space of
    the line. `dcc.Dropdown` automatically stretch with a minimum length
    (see `html_components.css`).

    Example:
        row = FlexRow(
            dcc.Dropown(["A", "B"]),
            html.Button("Accept"),
            id="example",
        )
    """

    def __init__(self, *children, className: Optional[str] = None, **kwargs):
        if className is None:
            className = "flex-row"
        else:
            className = className + " flex-row"
        super().__init__(
            children=list(children), className=className, **kwargs
        )


class InputText(html.Div):
    def __init__(
        self,
        *args,
        className: Optional[str] = None,
        type: str = "text",
        div_kws: Dict[str, Any] = {},
        **kwargs,
    ):
        """Create a `dcc.Input` wrapped in a `html.Div` with `className`s to
        make it match `dcc.Dropdown`.

        Args:
            *args: Arguments forwarded to `dcc.Input`.
            className: Additional classnames for the `html.Div`. Defaults to
                None.
            type: Type of `dcc.Input`. Defaults to "text".
            div_kws: Additional arguments to the `html.Div`. Defaults to {}.
            **kwargs: Keyword arguments forwarded to `dcc.Input`.
        """
        inp = dcc.Input(
            *args, type=type, className="Select-input Select-control", **kwargs
        )
        if className is not None:
            className = "Select " + className
        else:
            className = "Select"
        super().__init__(children=inp, className=className, **div_kws)


class ColumnDropdown(dcc.Dropdown):
    def __init__(
        self,
        id: Dict,
        *args,
        **kwargs,
    ):
        """Create a `dcc.Dropdown` for selecting columns from the data frame
        and auxiliary data frame.

        Args:
            id: Element id.
            *args: Arguments forwarded to `dcc.Dropdown`.
            regex: Add a hidden synced input field for regex matching.
            *kwargs: Keyword arguments forwarded to `dcc.Dropdown`.
        """
        super().__init__(
            *args,
            id=id,
            **kwargs,
        )

    @classmethod
    def get_columns(
        cls,
        df: pd.DataFrame,
        aux: Optional[pd.DataFrame] = None,
        options: List[str] = [],
        numeric: bool = False,
        category: bool = False,
    ) -> List[str]:
        """Get column names from the data frame and auxiliary data frame.

        Args:
            df: Data frame.
            aux: Auxiliary data frame. Defaults to None.
            options: Additional options to add to the column names. Defaults to [].
            numeric: Only select numeric columns. Defaults to False.
            category: Only select categorical (and integer, and object) columns.
              Defaults to False.

        Returns:
            List of column names
        """
        if aux is None:
            aux = pd.DataFrame()
        if numeric:
            return (
                options
                + df.select_dtypes("number").columns.to_list()
                + aux.select_dtypes("number").columns.to_list()
            )
        elif category:
            return (
                options
                + df.select_dtypes(
                    ["number", "category", "object"], "float"
                ).columns.to_list()
                + aux.select_dtypes(
                    ["number", "category", "object"], "float"
                ).columns.to_list()
            )
        else:
            return options + df.columns.to_list() + aux.columns.to_list()

    @classmethod
    def register_callback(
        cls,
        app: Dash,
        id: Dict,
        df_from_store: Callable,
        options: List[str] = [],
        numeric: bool = False,
        category: bool = False,
        regex_button_id: Optional[Dict] = None,
        regex_input_id: Optional[Dict] = None,
    ):
        """Register callbacks for the dropdowns.
        Call this once per unique id (ignoring index).

        Args:
            app: Dash app.
            id: Dropdown id.
            df_from_store: Function for extracting dataframes from store.
            options: See `ColumnDropdown.get_columns()`. Defaults to [].
            numeric: See `ColumnDropdown.get_columns()`. Defaults to False.
            category: See `ColumnDropdown.get_columns()`. Defaults to False.
            regex_button_id: Id for the "add regex" button, if this dropdown
              accepts regex. Defaults to None.
            regex_input_id: Id for a hidden input (that will be synced to the
              dropdown), if this dropdown accepts regex. Defaults to None.
        """
        if regex_button_id is None or regex_input_id is None:
            if isinstance(id, dict) and "index" in id:
                id["index"] = ALL

            @app.callback(
                Output(id, "options"),
                Input("data_frame_store", "data"),
                Input("auxiliary_store", "data"),
                State(id, "value"),
                prevent_initial_call=False,
            )
            def update_columns(df, aux, dds):
                if df is None:
                    return no_update
                df = df_from_store(df)
                aux = decode_aux(aux)
                columns = cls.get_columns(df, aux, options, numeric, category)
                if isinstance(id, dict) and "index" in id:
                    return [columns] * len(dds)
                else:
                    return columns

        else:
            if isinstance(id, dict) and "index" in id:
                id["index"] = MATCH
                regex_button_id["index"] = MATCH
                regex_input_id["index"] = MATCH

            # Sync `dropdown.search_value` and "regex_input"
            app.clientside_callback(
                """
                function (input) {
                if (input == "") {
                    return window.dash_clientside.no_update;
                }
                return input;
                }
                """,
                Output(regex_input_id, "value"),
                Input(id, "search_value"),
            )

            @app.callback(
                Output(id, "options"),
                Output(id, "value"),
                Output(id, "search_value"),
                Input("data_frame_store", "data"),
                Input("auxiliary_store", "data"),
                Input(regex_button_id, "n_clicks"),
                Input(id, "value"),
                State(regex_input_id, "value"),
                prevent_initial_call=False,
            )
            def update_columns_regex(df, aux, _nc, selected, regex):
                if df is None:
                    return no_update, no_update, no_update
                df = df_from_store(df)
                aux = decode_aux(aux)
                columns = cls.get_columns(df, aux, options, numeric, category)

                if ctx.triggered_id == regex_button_id or (
                    isinstance(ctx.triggered_id, dict)
                    and ctx.triggered_id["type"] == regex_button_id["type"]
                ):
                    new_search = ""
                else:
                    regex = None
                    new_search = no_update

                columns, selected, _ = dropdown_regex(columns, selected, regex)
                return columns, selected, new_search


class ClusterDropdown(dcc.Dropdown):
    def __init__(
        self,
        *args,
        index: Optional[str] = None,
        id: Optional[Dict] = None,
        **kwargs,
    ):
        """Create a `dcc.Dropdown` for selecting clusters.

        Args:
            index: Index of the dropdown (to be used with the default id).
              Defaults to None.
            id: Optionally override the default id (see
              `ClusterDropdown.register_callbacks()`). Defaults to None.
        """
        if not id:
            id = self.get_id(index if index is not None else str(uuid.uuid4()))
        super().__init__(*args, id=id, options=self.get_options(), **kwargs)

    @classmethod
    def get_id(cls, index: str):
        return generate_id(cls, index)

    @classmethod
    def get_options(
        cls, clusters: Optional[pd.Categorical] = None, empty: bool = False
    ) -> List[object]:
        """Generate the options for a ClusterDropdown.

        Args:
            clusters: Optional column of cluster ids. Defaults to None.
            empty: Include emtpy clusters. Defaults to False.

        Returns:
            List of Dropdown options.
        """
        cc = cluster_colours()
        if clusters is None:
            counter = {k: "" for k in cc.keys()}
        else:
            counter = {k: f": [{v}]" for k, v in Counter(clusters).items()}
            counter["all"] = f": [{len(clusters)}]"
        if empty:
            for k in cc.keys():
                if k not in counter:
                    counter[k] = ""

        options = [
            {
                "label": html.Div(
                    [
                        html.Div(
                            style={"background-color": colour},
                            className="color-rect",
                        ),
                        html.Div(
                            (
                                f"Everything{counter[cluster]}"
                                if cluster == "all"
                                else f"Cluster #{i}{counter[cluster]}"
                            ),
                            style={
                                "display": "inline-block",
                                "padding-left": 10,
                            },
                        ),
                    ],
                    style={"display": "flex", "align-items": "center"},
                ),
                "value": cluster,
            }
            for i, (cluster, colour) in enumerate(cluster_colours().items())
            if cluster in counter
        ]
        return options

    @classmethod
    def register_callbacks(
        cls,
        app: Dash,
        id: Union[None, str, Dict] = None,
        empty: bool = False,
    ):
        """Register callbacks for the cluster dropdowns.
        This is only needed if the id is changed (otherwise it is already setup).

        Args:
            app: Dash app
            id: Custom id (dicts should have an `ALL` component). Defaults to None.
            empty: Include empty clausers. Defaults to False.
        """
        if id is None:
            id = cls.get_id(ALL)

        @app.callback(
            Output(id, "options"),
            Input("auxiliary_store", "data"),
            prevent_initial_call=False,
        )
        def cluster_dropdown_count_callback(aux):
            if aux is None:
                return no_update
            clusters = get_clusters(aux)
            options = cls.get_options(clusters, empty)
            if isinstance(id, dict):
                return [options] * len(ctx.outputs_list)
            else:
                return options


class DeleteButton(html.Button):
    def __init__(self, index: Any, children: str = "x", **kwargs: Any):
        """Create a delete button.

        Args:
            index: Which index should be deleted.
            children: The visuals of the button. Defaults to "x".
            **kwargs: additional arguments forwarded to `html.Button`.
        """
        super().__init__(
            children=children,
            id=generate_id(DeleteButton, index),
            className="delete",
            **kwargs,
        )


try:
    import kaleido  # noqa: F401

    class PdfButton(html.Button):
        def __init__(
            self,
            index: str,
            plot_name: str,
            children: str = "Download as pdf",
            className: str = "pdf_button",
            **kwargs: Any,
        ):
            """Create a button for donwloading plots as pdf.

            Args:
                index: Which plot should be downloaded.
                children: The text of the button. Defaults to
                    "Download as pdf".
                **kwargs: additional arguments forwarded to `html.Button`.
            """
            id = generate_id(type(self), index)
            id["plot"] = plot_name
            id2 = generate_id(type(self), index, "download")
            id2["plot"] = plot_name
            children = [children, dcc.Download(id=id2)]
            super().__init__(
                children=children, id=id, className=className, **kwargs
            )

        @classmethod
        def register_callback(
            cls, app: Dash, plot_name: str, graph_id: Dict[str, str]
        ):
            """Register callbacks.
            NOTE: Currently this method has to be called separately for every
            type of graph.
            This is because Dash cannot match based on properties (i.e.,
            select only the `State`s with a "figure" property).

            Args:
                app: Xiplot app.
                plot_name: Name of the plot.
                graph_id: Id of the graph.
            """
            id = generate_id(cls, MATCH)
            id["plot"] = plot_name
            id2 = generate_id(cls, MATCH, "download")
            id2["plot"] = plot_name
            graph_id["index"] = MATCH

            @app.callback(
                Output(id2, "data"),
                Input(id, "n_clicks"),
                State(graph_id, "figure"),
                prevent_initial_call=True,
            )
            def download_as_pdf(n_clicks, figure):
                if not figure:
                    return no_update
                fig_img = po.io.to_image(figure, format="pdf")
                file = BytesIO(fig_img)
                encoded = base64.b64encode(file.getvalue()).decode("ascii")
                return dict(
                    base64=True,
                    content=encoded,
                    filename="xiplot.pdf",
                    type="application/pdf",
                )

except ImportError:

    class PdfButton(html.Div):
        def __init__(
            self,
            index: str,
            plot_name: str,
            children: str = "Download as pdf",
            **kwargs: Any,
        ):
            super().__init__(id=generate_id(type(self), index), **kwargs)

        @classmethod
        def register_callback(
            cls, app: Dash, plot_name: str, graph_id: Dict[str, str]
        ):
            pass


class PlotData(dcc.Store):
    @classmethod
    def get_id(cls, index: Any, plot_name: str) -> Dict[str, Any]:
        id = generate_id(cls, index)
        id["plot"] = plot_name
        return id

    def __init__(self, index: Any, plot_name: str, **kwargs):
        """
        Create a `dcc.Store` for storing plot configurations.
        These are used to save the xiplot state to a file.

        Args:
            index: The plot index.
            plot_name: The plot name (`APlot.name()`).
        """
        super().__init__(
            id=self.get_id(index, plot_name),
            data=dict(index=index, type=plot_name),
            **kwargs,
        )

    @classmethod
    def register_callback(
        cls,
        plot_name: str,
        app: Dash,
        inputs: Union[Dict[str, Union[Input, State]], Any],
        process: Optional[Callable[[Any], Dict[str, Any]]] = None,
    ):
        """Register callbacks for updating whenever the configuration changes.
        NOTE: This method has to be called separately for every type of plot.

        Args:
            plot_name: The plot name (`APlot.name()`).
            app: The xiplot app.
            inputs: Dictionary of `Input`s that control the plot. Note that
                index must be `MATCH`. Example: ```
                dict(
                    var=Input({
                        "type": "var-dropdown", "index": MATCH
                    }, "value")
                )
                ```
            process: Optional function that processes the `inputs` into a
                storable dictionary. Defaults to None.
        """
        # This assert does not verify that 'process' returns a dict
        assert callable(process) or isinstance(inputs, dict), (
            "'inputs' must be a `dict` or 'process' must be a function"
            " returning a `dict`"
        )

        # Try to enforce MATCH on the "index"
        if isinstance(inputs, dict):
            for v in inputs.values():
                v.component_id["index"] = MATCH
        elif isinstance(inputs, Sequence):
            for v in inputs:
                v.component_id["index"] = MATCH
        elif isinstance(inputs, Input):
            inputs.component_id["index"] = MATCH

        id = cls.get_id(MATCH, plot_name)

        @app.callback(
            Output(id, "data"),
            inputs={"meta": State(id, "data"), "inputs": inputs},
            prevent_initial_call=False,
        )
        def update_metadata(meta, inputs):
            if callable(process):
                inputs = process(inputs)
            for key, value in inputs.items():
                meta[key] = value
            return meta


class HelpButton(dcc.ConfirmDialogProvider):
    def __init__(self, text: str, btn_kws: Dict[str, Any] = {}, **kwargs):
        """Create a button that shows a text on hover, and the same text in a
        dialog when clicked.

        Args:
            text: The (help) text to be shown.
            btn_kws: Keyword arguments for the button.
            **kwargs: Keyword arguments for the wrapper.
        """
        super().__init__(
            html.Button("?", title=text, **btn_kws), message=text, **kwargs
        )



import dash
import plotly.express as px
from dash import ALL, MATCH, Input, Output, State, dcc
from dash.exceptions import PreventUpdate

from xiplot.plots import APlot
from xiplot.plugin import (
    ID_AUXILIARY,
    ID_CLICKED,
    ID_HOVERED,
    placeholder_figure,
)
from xiplot.utils.auxiliary import (
    CLUSTER_COLUMN_NAME,
    SELECTED_COLUMN_NAME,
    decode_aux,
    merge_df_aux,
    toggle_selected,
)
from xiplot.utils.cluster import cluster_colours
from xiplot.utils.components import (
    ColumnDropdown,
    FlexRow,
    PdfButton,
    PlotData,
)
from xiplot.utils.dataframe import get_default_column
from xiplot.utils.layouts import layout_wrapper


class Lineplot(APlot):
    @classmethod
    def register_callbacks(cls, app, df_from_store, df_to_store):
        PdfButton.register_callback(app, cls.name(), cls.get_id(MATCH))

        @app.callback(
            Output(cls.get_id(MATCH), "figure"),
            Input(cls.get_id(MATCH, "x_axis"), "value"),
            Input(cls.get_id(MATCH, "y_axis"), "value"),
            Input(cls.get_id(MATCH, "color"), "value"),
            Input(ID_HOVERED, "data"),
            Input("data_frame_store", "data"),
            Input("auxiliary_store", "data"),
            Input("plotly-template", "data"),
            prevent_initial_call=False,
        )
        def render(
            x_axis,
            y_axis,
            color,
            hover,
            df,
            aux,
            template=None,
        ):
            return cls.render(
                df_from_store(df),
                decode_aux(aux),
                x_axis,
                y_axis,
                color,
                hover,
                template,
            )

        def get_row(hover):
            try:
                for p in hover:
                    if p is not None:
                        return p["points"][0]["customdata"][0]
            except (KeyError, TypeError):
                raise PreventUpdate()
            raise PreventUpdate()

        @app.callback(
            Output(ID_HOVERED, "data"),
            Output(cls.get_id(ALL), "hoverData"),
            Input(cls.get_id(ALL), "hoverData"),
        )
        def handle_hover_events(hover):
            return get_row(hover), [None] * len(hover)

        @app.callback(
            Output(ID_AUXILIARY, "data"),
            Output(ID_CLICKED, "data"),
            Output(cls.get_id(ALL), "clickData"),
            Input(cls.get_id(ALL), "clickData"),
            State(ID_AUXILIARY, "data"),
        )
        def handle_click_events(click, aux):
            row = get_row(click)
            if aux is None:
                return dash.no_update, row
            return toggle_selected(aux, row), row, [None] * len(click)

        PlotData.register_callback(
            cls.name(),
            app,
            {
                "x_axis": Input(cls.get_id(MATCH, "x_axis"), "value"),
                "y_axis": Input(cls.get_id(MATCH, "y_axis"), "value"),
                "color": Input(cls.get_id(MATCH, "color"), "value"),
            },
        )

        ColumnDropdown.register_callback(
            app, cls.get_id(ALL, "x_axis"), df_from_store, numeric=True
        )
        ColumnDropdown.register_callback(
            app, cls.get_id(ALL, "y_axis"), df_from_store, numeric=True
        )
        ColumnDropdown.register_callback(
            app, cls.get_id(ALL, "color"), df_from_store, category=True
        )

        return render, handle_hover_events, handle_click_events

    @staticmethod
    def render(
        df,
        aux,
        x_axis,
        y_axis,
        color=None,
        hover=None,
        template=None,
    ):
        df = merge_df_aux(df, aux)
        df["__Xiplot_index__"] = range(df.shape[0])
        if x_axis not in df.columns or y_axis not in df.columns:
            return placeholder_figure("Please select x and y axis")
        if color not in df.columns:
            color = None
        fig = px.line(
            df.sort_values(by=x_axis),
            x=x_axis,
            y=y_axis,
            color=color,
            color_discrete_map=cluster_colours(),
            custom_data=["__Xiplot_index__"],
            template=template,
        )
        if hover is not None:
            fig.add_vline(
                df[x_axis][hover],
                line=dict(color="rgba(0.5,0.5,0.5,0.5)", dash="dash"),
                layer="below",
            )
            fig.add_hline(
                df[y_axis][hover],
                line=dict(color="rgba(0.5,0.5,0.5,0.5)", dash="dash"),
                layer="below",
            )
        if SELECTED_COLUMN_NAME in aux:
            df2 = df[aux[SELECTED_COLUMN_NAME]]
            if not df2.empty:
                trace = px.scatter(df2, x=x_axis, y=y_axis)
                color = "#DDD" if template and "dark" in template else "#333"
                trace.update_traces(
                    hoverinfo="skip",
                    hovertemplate=None,
                    marker=dict(size=15, color=color),
                )
                fig.add_traces(trace.data)
        return fig

    @classmethod
    def create_layout(cls, index, df, columns=None, config=dict()):
        num_columns = ColumnDropdown.get_columns(df, numeric=True)
        cat_columns = ColumnDropdown.get_columns(df, category=True)

        x_axis = config.get("x_axis", get_default_column(num_columns, "x"))
        y_axis = config.get("y_axis", get_default_column(num_columns, "y"))
        color = config.get("color", CLUSTER_COLUMN_NAME)

        return [
            dcc.Graph(id=cls.get_id(index)),
            FlexRow(
                layout_wrapper(
                    component=ColumnDropdown(
                        cls.get_id(index, "x_axis"),
                        options=num_columns,
                        value=x_axis,
                        clearable=False,
                    ),
                    css_class="dash-dropdown",
                    title="X",
                ),
                layout_wrapper(
                    component=ColumnDropdown(
                        cls.get_id(index, "y_axis"),
                        options=num_columns,
                        value=y_axis,
                        clearable=False,
                    ),
                    css_class="dash-dropdown",
                    title="Y",
                ),
                layout_wrapper(
                    component=ColumnDropdown(
                        cls.get_id(index, "color"),
                        options=cat_columns,
                        value=color,
                        clearable=True,
                    ),
                    css_class="dash-dropdown",
                    title="Groups",
                ),
            ),
        ]



def get_sort_by(sort_by, selected_rows, trigger):
    all_selected = all(selected_rows) or not any(selected_rows)
    if len(sort_by) == 0 and all_selected:
        sort_by = []

    elif len(sort_by) == 0 and not all_selected:
        sort_by = [
            {"column_id": "Selection", "direction": "desc"},
            {"column_id": "index_copy", "direction": "asc"},
        ]

    elif len(sort_by) != 0 and True not in selected_rows:
        pass

    elif sort_by[0]["column_id"] != "Selection":
        sort_by.insert(0, {"column_id": "Selection", "direction": "desc"})

    elif (
        trigger != "auxiliary_store"
        and {"column_id": "index_copy", "direction": "asc"} in sort_by
    ):
        sort_by.remove({"column_id": "index_copy", "direction": "asc"})
        sort_by.append({"column_id": "index_copy", "direction": "asc"})

    elif (
        all_selected
        and {"column_id": "index_copy", "direction": "desc"} in sort_by
    ):
        sort_by.remove({"column_id": "Selection", "direction": "desc"})
        sort_by.remove({"column_id": "index_copy", "direction": "asc"})

    elif sort_by == {"column_id": "Selection", "direction": "desc"}:
        sort_by.append({"column_id": "index_copy", "direction": "asc"})

    elif len(sort_by) == 1 and sort_by[0]["column_id"] == "Selection":
        sort_by.append({"column_id": "index_copy", "direction": "asc"})

    return sort_by


def get_updated_item(items, index, inputs_list):
    """
    Return the of the item that has been updated among all items except a new
    item entry

    Parameters:
        items: list of items (Input ALL gets list of all items that has same
            id type)
        index: index of the item that has been updated
        inputs_list: list of all inputs of the callback

    Returns:
        item: item that has been updated
    """
    # Ignore a new item
    if items[-1] is None:
        items = items[:-1]

    # Convert all None values to an empty list
    for id, item in enumerate(items):
        if item is None:
            items[id] = []

    # Get the updated item
    item_id = get_updated_item_id(items, index, inputs_list)
    item = items[item_id]

    return item


def get_updated_item_id(items, index, inputs_list):
    for id, item in enumerate(inputs_list):
        if item["id"]["index"] == index:
            return id



"""
    This module exposes the parts of xiplot that are useful for implementing plugins.
    Most of the members of this module are just imported from other places in xiplot.
"""

from io import BytesIO
from os import PathLike
from typing import Any, Callable, Dict, Optional, Tuple, Union

import pandas as pd
from dash import Dash
from dash.development.base_component import Component

# Export useful parts of xiplot:
from xiplot.plots import APlot  # noqa: F401
from xiplot.utils import generate_id  # noqa: F401
from xiplot.utils.auxiliary import (  # noqa: F401
    decode_aux,
    encode_aux,
    get_clusters,
    get_selected,
    merge_df_aux,
)
from xiplot.utils.cluster import cluster_colours  # noqa: F401
from xiplot.utils.components import (  # noqa: F401
    ClusterDropdown,
    ColumnDropdown,
    DeleteButton,
    FlexRow,
    HelpButton,
    PdfButton,
    PlotData,
)

# IDs for important `dcc.Store` components:
ID_DATAFRAME = "data_frame_store"  # Main dataframe (readonly)
ID_AUXILIARY = "auxiliary_store"  # Additional dataframe (editable)
ID_METADATA = "metadata_store"
ID_HOVERED = "lastly_hovered_point_store"
ID_CLICKED = "lastly_clicked_point_store"
# `dcc.Store` that is `True` if the dark mode is active
ID_DARK_MODE = "light-dark-toggle-store"
# `dcc.Store` that contains the current plotly template (used for the dark mode)
ID_PLOTLY_TEMPLATE = "plotly-template"

# Useful CSS classes
CSS_STRETCH_CLASS = "stretch"
CSS_LARGE_BUTTON_CLASS = "button"
CSS_DELETE_BUTTON_CLASS = "delete"


# Helpful typehints:
AReadFunction = Callable[[Union[BytesIO, PathLike]], pd.DataFrame]
# Read plugin return string: file extension
AReadPlugin = Callable[[], Optional[Tuple[AReadFunction, str]]]
AWriteFunction = Callable[[pd.DataFrame, BytesIO], None]
# Write plugin return strings: file extension, MIME type
#  (ususally "application/octet-stream")
AWritePlugin = Callable[[], Optional[Tuple[AWriteFunction, str, str]]]
AGlobalPlugin = Callable[[], Component]
# Callback plugin functions: parse_to_dataframe, serialise_from_dataframe
ACallbackPlugin = Callable[
    [Dash, Callable[[Any], pd.DataFrame], Callable[[pd.DataFrame], Any]], None
]


# Helper functions:
def placeholder_figure(text: str) -> Dict[str, Any]:
    """Display a placeholder text instead of a graph.
    This can be used in a "callback" function when a graph cannot be rendered.

    Args:
        text: Placeholder text.

    Returns:
        Dash figure (to place into a `Output(dcc.Graph.id, "figure")`).
    """
    return {
        "layout": {
            "xaxis": {"visible": False},
            "yaxis": {"visible": False},
            "annotations": [
                {
                    "text": text,
                    "xref": "paper",
                    "yref": "paper",
                    "showarrow": False,
                    "font": {"size": 28},
                }
            ],
        }
    }



import time

import dash
import pandas as pd
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from tests.util_test import render_plot, start_server
from xiplot.plots.histogram import Histogram

render = Histogram.register_callbacks(
    dash.Dash(__name__), lambda x: x, lambda x: x
)[0]


def test_tehi001_render_histogram(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Histogram")

    plot = driver.find_element(By.CLASS_NAME, "dash-graph")

    assert "histogram" in plot.get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_tehi002_set_axis(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Histogram")

    driver.find_element(By.CLASS_NAME, "dd-single").click()

    x = driver.find_element(
        By.XPATH,
        "//div[@class='dd-single']/div[2]/div[1]/div[1]/div[1]/div[2]/input",
    )

    x.send_keys("mpg")
    x.send_keys(Keys.RETURN)
    time.sleep(0.5)

    assert "mpg" in driver.find_element(By.CLASS_NAME, "xtitle").text
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_tehi003_clear_clusters(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Histogram")

    cluster_dd = driver.find_element(
        By.XPATH,
        "//div[@class='plots']/div[3]/div[2]/div[1]/div[1]/span[1]",
    )
    cluster_dd.click()

    assert "Select..." in driver.find_element(
        By.XPATH,
        "//div[@class='plots']/div[4]/div[2]/div[1]/div[1]",
    ).get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_create_histogram():
    df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    output = render("col1", "all", 1, df, pd.DataFrame())
    fig = output

    assert str(type(fig)) == "<class 'plotly.graph_objs._figure.Figure'>"



import dash
import plotly.express as px
from dash import ALL, MATCH, Input, Output, State, dcc
from dash.exceptions import PreventUpdate

from xiplot.plots import APlot
from xiplot.plugin import (
    ID_AUXILIARY,
    ID_CLICKED,
    ID_HOVERED,
    placeholder_figure,
)
from xiplot.utils.auxiliary import (
    CLUSTER_COLUMN_NAME,
    SELECTED_COLUMN_NAME,
    decode_aux,
    merge_df_aux,
    toggle_selected,
)
from xiplot.utils.cluster import cluster_colours
from xiplot.utils.components import (
    ColumnDropdown,
    FlexRow,
    PdfButton,
    PlotData,
)
from xiplot.utils.layouts import layout_wrapper


class Distplot(APlot):
    @classmethod
    def name(cls):
        return "Density plot"

    @classmethod
    def register_callbacks(cls, app, df_from_store, df_to_store):
        PdfButton.register_callback(app, cls.name(), cls.get_id(MATCH))

        @app.callback(
            Output(cls.get_id(MATCH), "figure"),
            Input(cls.get_id(MATCH, "variable"), "value"),
            Input(cls.get_id(MATCH, "color"), "value"),
            Input(ID_HOVERED, "data"),
            Input("data_frame_store", "data"),
            Input("auxiliary_store", "data"),
            Input("plotly-template", "data"),
            prevent_initial_call=False,
        )
        def render(
            variable,
            color,
            hover,
            df,
            aux,
            template=None,
        ):
            return cls.render(
                df_from_store(df),
                decode_aux(aux),
                variable,
                color,
                hover,
                template,
            )

        def get_row(hover):
            try:
                for p in hover:
                    if p is not None:
                        return p["points"][0]["customdata"]
            except (KeyError, TypeError):
                raise PreventUpdate()
            raise PreventUpdate()

        @app.callback(
            Output(ID_HOVERED, "data"),
            Output(cls.get_id(ALL), "hoverData"),
            Input(cls.get_id(ALL), "hoverData"),
        )
        def handle_hover_events(hover):
            return get_row(hover), [None] * len(hover)

        @app.callback(
            Output(ID_AUXILIARY, "data"),
            Output(ID_CLICKED, "data"),
            Output(cls.get_id(ALL), "clickData"),
            Input(cls.get_id(ALL), "clickData"),
            State(ID_AUXILIARY, "data"),
        )
        def handle_click_events(click, aux):
            row = get_row(click)
            if aux is None:
                return dash.no_update, row
            return toggle_selected(aux, row), row, [None] * len(click)

        PlotData.register_callback(
            cls.name(),
            app,
            {
                "variable": Input(cls.get_id(MATCH, "variable"), "value"),
                "color": Input(cls.get_id(MATCH, "color"), "value"),
            },
        )

        ColumnDropdown.register_callback(
            app, cls.get_id(ALL, "variable"), df_from_store, numeric=True
        )
        ColumnDropdown.register_callback(
            app, cls.get_id(ALL, "color"), df_from_store, category=True
        )

        return render, handle_hover_events, handle_click_events

    @staticmethod
    def render(
        df,
        aux,
        variable,
        color=None,
        hover=None,
        template=None,
    ):
        # `plotly.figure_factory` checks if scipy is installed and caches the
        # result. This causes issues if scipy is lazily loaded in WASM.
        import scipy  # noqa: F401, isort: skip
        import plotly.figure_factory as ff

        df = merge_df_aux(df, aux)
        df["__Xiplot_index__"] = range(df.shape[0])
        if variable not in df.columns:
            return placeholder_figure("Please select a variable")
        if color not in df.columns:
            fig = ff.create_distplot(
                [df[variable]], [variable], show_hist=False
            )
            fig.update_layout(
                template=template,
                showlegend=False,
                xaxis_title=variable,
                yaxis_title="Density",
                yaxis_domain=[0.11, 1],
                yaxis2_domain=[0, 0.09],
            )
            fig.data[1]["customdata"] = df["__Xiplot_index__"]

        else:
            df[color] = df[color].astype("category")
            df2 = df.groupby(color)
            cats = [
                c
                for c in df[color].cat.categories
                if df2[variable].get_group(c).std() > 1e-6
            ]
            data = [df2[variable].get_group(g) for g in cats]
            customdata = [df2["__Xiplot_index__"].get_group(g) for g in cats]
            groups = [str(g) for g in cats]
            colors1 = px.colors.qualitative.Plotly
            colors2 = cluster_colours()
            colors = []
            for i, n in enumerate(groups):
                colors.append(colors2.get(n, colors1[i % len(colors1)]))
            fig = ff.create_distplot(
                data, groups, show_hist=False, colors=colors
            )
            fig.update_layout(
                template=template,
                legend=dict(title=color, traceorder="normal"),
                xaxis_title=variable,
                yaxis_title="Density",
                yaxis_domain=[0.16, 1] if len(cats) < 4 else [0.26, 1],
                yaxis2_domain=[0, 0.14] if len(cats) < 4 else [0, 0.24],
            )
            for i, d in enumerate(customdata):
                fig.data[i + len(cats)]["customdata"] = d

        if hover is not None:
            fig.add_vline(
                df[variable][hover],
                line=dict(color="rgba(0.5,0.5,0.5,0.5)", dash="dash"),
                layer="below",
            )
        if SELECTED_COLUMN_NAME in aux:
            sel_color = "#DDD" if template and "dark" in template else "#333"
            for x in df[variable][aux[SELECTED_COLUMN_NAME]]:
                fig.add_vline(
                    x, line=dict(color=sel_color, width=0.5), layer="below"
                )
            x = df[variable][aux[SELECTED_COLUMN_NAME]]
            fig.add_scatter(
                x=x,
                y=(
                    [variable] * len(x)
                    if color not in df.columns
                    else df[color][aux[SELECTED_COLUMN_NAME]]
                ),
                marker=dict(color=sel_color, size=8),
                yaxis="y2",
                hoverinfo="skip",
                hovertemplate=None,
                mode="markers",
                showlegend=False,
            )
        return fig

    @classmethod
    def create_layout(cls, index, df, columns=None, config=dict()):
        num_columns = ColumnDropdown.get_columns(df, numeric=True)

        variable = config.get(
            "variable", num_columns[0] if num_columns else None
        )
        color = config.get("color", CLUSTER_COLUMN_NAME)

        return [
            dcc.Graph(id=cls.get_id(index)),
            FlexRow(
                layout_wrapper(
                    component=ColumnDropdown(
                        cls.get_id(index, "variable"),
                        options=num_columns,
                        value=variable,
                        clearable=False,
                    ),
                    css_class="dash-dropdown",
                    title="Variable",
                ),
                layout_wrapper(
                    component=ColumnDropdown(
                        cls.get_id(index, "color"),
                        options=columns,
                        value=color,
                        clearable=True,
                    ),
                    css_class="dash-dropdown",
                    title="Groups",
                ),
            ),
        ]



import json
import tarfile
from collections import OrderedDict
from io import BytesIO
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
)

import pandas as pd

from xiplot.tabs.plugins import get_plugins_cached
from xiplot.utils.io import FinallyCloseBytesIO


def get_data_filepaths(data_dir=""):
    try:
        return sorted(
            (fp for fp in Path(data_dir).iterdir() if fp.is_file()),
            reverse=True,
        )

    except FileNotFoundError:
        return []


def read_functions() -> (
    Iterator[Tuple[Callable[[BytesIO], pd.DataFrame], str]]
):
    """Generate all functions for reading to a dataframe.

    Yields:
        fn: Function that reads the data and returns a dataframe.
        ext: File extension that the readed can handle.
    """
    for _, _, plugin in get_plugins_cached("read"):
        plugin = plugin()
        if plugin is not None:
            yield plugin

    def read_json(data):
        if isinstance(data, BytesIO):
            data = data.getvalue().decode("utf-8")
        elif isinstance(data, tarfile.ExFileObject):
            data = data.read().decode("utf-8")

        try:
            return pd.read_json(data, typ="frame", orient="columns")
        except Exception:
            return pd.read_json(data, typ="frame", orient="split")

    yield pd.read_csv, ".csv"
    yield read_json, ".json"


def write_functions() -> (
    Iterator[Tuple[Callable[[pd.DataFrame, BytesIO], None], str, str]]
):
    """Generate all functions for writing dataframes.

    Yields:
        fn: Function that writes the dataframe to bytes.
        ext: File extension that matches the written data.
        mime: MIME type of the written data.
    """
    for _, _, plugin in get_plugins_cached("write"):
        plugin = plugin()
        if plugin is not None:
            yield plugin

    yield lambda df, file: df.to_csv(file, index=False), ".csv", "text/csv"
    yield lambda df, file: df.to_json(
        file, orient="split", index=False
    ), ".json", "application/json"


def read_dataframe_with_extension(data, filename=None):
    """
    Read the given data and convert it to a pandas data frame

    Parameters:

        data: File name or File-like object
        filename: File name as a string

    Returns:

        df: Pandas data frame
        aux: Pandas data frame
        meta: dictionary of metadata
    """
    if filename is None:
        filename = data

    if ".tar" in Path(filename).suffixes:
        stem = Path(filename).name[: -len("".join(Path(filename).suffixes))]

        with (
            tarfile.open(name=data)
            if isinstance(data, (str, Path))
            else tarfile.open(fileobj=data)
        ) as tar:
            df_file = None
            df_name = None
            aux_file = None
            aux_name = None
            meta_file = None

            for member in tar.getmembers():
                if not member.isfile():
                    raise Exception(
                        f"Tar contains a non-file entry {member.name}"
                    )

                if Path(member.name).stem == "data":
                    if df_file is not None:
                        raise Exception("Tar contains more than one data file")
                    df_file = tar.extractfile(member)
                    df_name = Path(stem).with_suffix(Path(member.name).suffix)
                elif Path(member.name).stem == "aux":
                    if aux_file is not None:
                        raise Exception(
                            "Tar contains more than one auxiliary file"
                        )
                    aux_file = tar.extractfile(member)
                    aux_name = member.name
                elif member.name == "meta.json":
                    if meta_file is not None:
                        raise Exception("Tar contains more than metadata file")
                    meta_file = tar.extractfile(member)
                else:
                    raise Exception(
                        f"Tar contains extraneous file '{member.name}'"
                    )

            if df_file is None:
                raise Exception("Tar contains no data file called 'data.*'")

            if aux_file is None:
                raise Exception(
                    "Tar contains no auxiliary file called 'aux.*'"
                )

            if meta_file is None:
                raise Exception(
                    "Tar contains no metadata file called 'meta.json'"
                )

            metadata = (
                json.load(meta_file, object_pairs_hook=OrderedDict)
                or OrderedDict()
            )
            metadata["filename"] = str(df_name)

            df = read_only_dataframe(df_file, df_name)

            try:
                aux = read_only_dataframe(aux_file, aux_name)
            except pd.errors.EmptyDataError:
                aux = pd.DataFrame()

            if aux.empty:
                aux.index = df.index
            if df.shape[0] != aux.shape[0]:
                raise Exception(
                    "The dataframe and auxiliary data have different number of"
                    " rows."
                )

            return df, aux, metadata

    df = read_only_dataframe(data, filename)
    return (
        df,
        pd.DataFrame(index=df.index),
        OrderedDict(filename=str(filename)),
    )


def read_only_dataframe(data, filename):
    file_extension = Path(filename).suffix
    error = None

    for fn, ext in read_functions():
        if file_extension == ext:
            try:
                return fn(data)
            except Exception as e:
                error = e

    if error is not None:
        raise error
    raise Exception(f"Unsupported dataframe format '{file_extension}'")


def write_dataframe_and_metadata(
    df: pd.DataFrame,
    aux: pd.DataFrame,
    meta: Dict[str, Any],
    filepath: str,
    file,
    file_extension: Optional[str] = None,
) -> Tuple[str, str]:
    if not aux.empty and df.shape[0] != aux.shape[0]:
        raise Exception(
            "The dataframe and auxiliary data have different number of rows."
        )
    if file_extension is None:
        file_extension = Path(filepath).suffix

    with tarfile.open(fileobj=file, mode="w:gz") as tar:
        df_file = Path("data").with_suffix(file_extension).name
        aux_file = Path("aux").with_suffix(file_extension).name
        meta_file = "meta.json"
        tar_file = Path(filepath).with_suffix(".tar.gz").name

        with FinallyCloseBytesIO() as df_bytes:
            write_only_dataframe(df, filepath, df_bytes, file_extension)
            df_bytes = df_bytes.getvalue()

        df_info = tarfile.TarInfo(df_file)
        df_info.size = len(df_bytes)

        tar.addfile(df_info, BytesIO(df_bytes))

        with FinallyCloseBytesIO() as aux_bytes:
            write_only_dataframe(aux, aux_file, aux_bytes, file_extension)
            aux_bytes = aux_bytes.getvalue()

        aux_info = tarfile.TarInfo(aux_file)
        aux_info.size = len(aux_bytes)

        tar.addfile(aux_info, BytesIO(aux_bytes))

        meta = meta or OrderedDict()
        meta["filename"] = Path(filepath).name

        meta_string = json.dumps(meta)
        meta_bytes = meta_string.encode("utf-8")

        meta_info = tarfile.TarInfo(meta_file)
        meta_info.size = len(meta_bytes)

        tar.addfile(meta_info, BytesIO(meta_bytes))

        return tar_file, "application/gzip"


def write_only_dataframe(
    df: pd.DataFrame,
    filepath: str,
    file: BytesIO,
    file_extension: Optional[str] = None,
) -> Tuple[str, str]:
    if file_extension is None:
        file_name = Path(filepath).name
        file_extension = Path(filepath).suffix
    else:
        file_name = Path(filepath).with_suffix(file_extension).name
    error = None

    for fn, ext, mime in write_functions():
        if file_extension == ext:
            try:
                fn(df, file)
                return file_name, mime
            except Exception as e:
                error = e

    if error is not None:
        raise error
    raise Exception(f"Unsupported dataframe format '{file_extension}'")


def get_numeric_columns(df, columns=None):
    """
    Return only columns, which are numeric

    Parameters:
        df: pandas.DataFrame
        columns: columns of the data frame

    Returns:
        columns: numeric columns
    """
    if columns is not None:
        df = df[columns]
    return df.select_dtypes("number").columns.to_list()


def get_default_column(
    columns: List[str], axis: Literal["x", "y"]
) -> Optional[str]:
    """Get default values for x_axis and y_axis from a list of column names.

    Args:
        columns: List of column names.
        axis: Is this the "x" or "y" axis.

    Returns:
        Column name.
    """
    if len(columns) == 0:
        return None
    if axis == "x":
        for c in columns:
            if "x" in c or "1" in c or "X" in c:
                return c
        return columns[0]

    else:
        for c in columns:
            if "y" in c or "2" in c or "Y" in c:
                return c
        return columns[min(1, len(columns) - 1)]



import base64
import uuid
from collections import OrderedDict
from io import BytesIO
from pathlib import Path

import dash
import dash_mantine_components as dmc
from dash import ALL, Input, Output, State, ctx, dcc, html
from dash.exceptions import PreventUpdate
from dash_extensions.enrich import ServersideOutput

from xiplot.tabs import Tab
from xiplot.utils import generate_id
from xiplot.utils.auxiliary import (
    CLUSTER_COLUMN_NAME,
    SELECTED_COLUMN_NAME,
    decode_aux,
    encode_aux,
    get_clusters,
)
from xiplot.utils.cluster import cluster_colours
from xiplot.utils.components import FlexRow, PlotData
from xiplot.utils.dataframe import (
    get_data_filepaths,
    read_dataframe_with_extension,
    write_dataframe_and_metadata,
    write_functions,
    write_only_dataframe,
)
from xiplot.utils.io import FinallyCloseBytesIO
from xiplot.utils.layouts import layout_wrapper


class Data(Tab):
    @staticmethod
    def register_callbacks(app, df_from_store, df_to_store, data_dir=""):
        try:
            import dash_uploader as du

            @du.callback(
                output=[
                    ServersideOutput("uploaded_data_file_store", "data"),
                    ServersideOutput("uploaded_auxiliary_store", "data"),
                    Output("uploaded_metadata_store", "data"),
                    Output("data_files", "options"),
                    Output("data_files", "value"),
                    Output("file_uploader_container", "children"),
                    Output("data-tab-upload-notify-container", "children"),
                ],
                id="file_uploader",
            )
            def upload(status):
                upload_path = Path("uploads") / Path(status[0]).name

                uploader = du.Upload(
                    id="file_uploader",
                    text="Drag and Drop or Select a File to upload",
                    default_style={
                        "minHeight": 1,
                        "lineHeight": 4,
                        "height": "85px",
                    },
                )

                try:
                    df, aux, meta = read_dataframe_with_extension(
                        upload_path, upload_path.name
                    )
                except Exception as err:
                    return (
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                        [uploader],
                        dmc.Notification(
                            id=str(uuid.uuid4()),
                            color="yellow",
                            title="Warning",
                            message=[
                                html.Div(
                                    [
                                        f"The file {upload_path.name} ",
                                        html.I("(upload)"),
                                        (
                                            " could not be loaded as a data"
                                            f" frame: {err}."
                                        ),
                                    ]
                                )
                            ],
                            action="show",
                            autoClose=10000,
                        ),
                    )
                finally:
                    upload_path.unlink()

                return (
                    df_to_store(df),
                    encode_aux(aux),
                    meta,
                    generate_dataframe_options(upload_path, data_dir),
                    str(Path("uploads") / upload_path.name),
                    [uploader],
                    None,
                )

        except (ImportError, AttributeError):

            @app.callback(
                ServersideOutput("uploaded_data_file_store", "data"),
                ServersideOutput("uploaded_auxiliary_store", "data"),
                Output("uploaded_metadata_store", "data"),
                Output("data_files", "options"),
                Output("data_files", "value"),
                Output("file_uploader", "contents"),
                Output("file_uploader", "filename"),
                Output("data-tab-upload-notify-container", "children"),
                Input("file_uploader", "contents"),
                State("file_uploader", "filename"),
            )
            def upload(contents, upload_name):
                if contents is None:
                    raise PreventUpdate()

                upload_name = Path(upload_name)
                _content_type, content_string = contents.split(",")
                decoded = base64.b64decode(content_string)

                try:
                    df, aux, meta = read_dataframe_with_extension(
                        BytesIO(decoded), upload_name
                    )
                except Exception as err:
                    return (
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                        None,
                        None,
                        dmc.Notification(
                            id=str(uuid.uuid4()),
                            color="yellow",
                            title="Warning",
                            message=[
                                html.Div(
                                    [
                                        f"The file {upload_name} ",
                                        html.I("(upload)"),
                                        (
                                            " could not be loaded as a data"
                                            f" frame: {err}."
                                        ),
                                    ]
                                )
                            ],
                            action="show",
                            autoClose=10000,
                        ),
                    )

                return (
                    df_to_store(df),
                    encode_aux(aux),
                    meta,
                    generate_dataframe_options(upload_name, data_dir),
                    str(Path("uploads") / upload_name.name),
                    None,
                    None,
                    None,
                )

        @app.callback(
            ServersideOutput("data_frame_store", "data"),
            Output("auxiliary_store", "data"),
            Output("metadata_store", "data"),
            Output("data-tab-notify-container", "children"),
            Input("submit-button", "n_clicks"),
            Input("uploaded_data_file_store", "data"),
            Input("uploaded_auxiliary_store", "data"),
            Input("uploaded_metadata_store", "data"),
            State("data_files", "value"),
        )
        def choose_file(
            data_btn,
            uploaded_data,
            uploaded_aux,
            uploaded_meta,
            filepath,
        ):
            trigger = ctx.triggered_id

            if not filepath:
                return (
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message="You have not selected a data file to load.",
                        action="show",
                        autoClose=10000,
                    ),
                )

            filepath = Path(filepath)

            notification = None

            if trigger == "submit-button":
                if str(list(filepath.parents)[0]) == "uploads":
                    df_store = uploaded_data
                    aux_store = uploaded_aux
                    meta = uploaded_meta

                    notification = dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="green",
                        title="Success",
                        message=[
                            html.Div(
                                [
                                    f"The data file {filepath.name} ",
                                    html.I("(upload)"),
                                    " was loaded successfully!",
                                ]
                            )
                        ],
                        action="show",
                        autoClose=5000,
                    )
                else:
                    filepath = Path(data_dir) / filepath.name

                    try:
                        df, aux, meta = read_dataframe_with_extension(
                            filepath, filepath.name
                        )
                    except Exception as err:
                        return (
                            dash.no_update,
                            dash.no_update,
                            dash.no_update,
                            dash.no_update,
                            dmc.Notification(
                                id=str(uuid.uuid4()),
                                color="yellow",
                                title="Warning",
                                message=(
                                    f"The file {filepath.name} could not be"
                                    f" loaded as a data frame: {err}."
                                ),
                                action="show",
                                autoClose=10000,
                            ),
                        )

                    df_store = df_to_store(df)
                    aux_store = encode_aux(aux)

                    notification = dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="green",
                        title="Success",
                        message=(
                            f"The data file {meta['filename']} was loaded"
                            " successfully!"
                        ),
                        action="show",
                        autoClose=5000,
                    )
            elif trigger == "uploaded_data_file_store":
                df_store = uploaded_data
                aux_store = uploaded_aux
                meta = uploaded_meta

                notification = dmc.Notification(
                    id=str(uuid.uuid4()),
                    color="green",
                    title="Success",
                    message=[
                        html.Div(
                            [
                                f"The data file {filepath.name} ",
                                html.I("(upload)"),
                                " was uploaded successfully!",
                            ]
                        )
                    ],
                    action="show",
                    autoClose=5000,
                )

            if meta.get("settings") is None:
                meta["settings"] = dict()

            if meta.get("plots") is None:
                meta["plots"] = OrderedDict()
            meta["session"] = str(uuid.uuid4())

            import jsonschema

            try:
                jsonschema.validate(
                    instance=meta,
                    schema=dict(
                        type="object",
                        properties=dict(
                            filename=dict(type="string"),
                            settings=dict(type="object"),
                            plots=dict(type="object"),
                        ),
                        required=["filename", "settings", "plots"],
                    ),
                )
            except jsonschema.exceptions.ValidationError as err:
                return (
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message=(
                            f"The file {filepath.name} has invalid metadata at"
                            f" meta{err.json_path[1:]}: {err.message}."
                        ),
                        action="show",
                        autoClose=10000,
                    ),
                )

            df = df_from_store(df_store)
            aux = decode_aux(aux_store)

            if SELECTED_COLUMN_NAME in aux:
                if aux.dtypes[SELECTED_COLUMN_NAME] != bool:
                    notification = dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message=(
                            f"The file {filepath.name} has been loaded but the"
                            ' auxiliary column "{SELECTED_COLUMN_NAME}" is'
                            " not of type bool."
                        ),
                        action="show",
                        autoClose=10000,
                    )

            if CLUSTER_COLUMN_NAME in aux:
                invalid_clusters = set(get_clusters(aux).categories) - set(
                    cluster_colours().keys()
                )
                if len(invalid_clusters) > 0:
                    message = (
                        f"The file {filepath.name} has been loaded"
                        f' but the auxiliary column "{CLUSTER_COLUMN_NAME}"'
                        f" contains invalid values {invalid_clusters}."
                    )
                    notification = dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message=message,
                        action="show",
                        autoClose=10000,
                    )

            return (
                df_store,
                aux_store,
                meta,
                notification,
            )

        @app.callback(
            Output("data-download", "data"),
            Output("data-tab-download-notify-container", "children"),
            Input("download-data-file-button", "n_clicks"),
            Input("download-plots-file-button", "n_clicks"),
            State("data_files", "value"),
            State("data_frame_store", "data"),
            State("auxiliary_store", "data"),
            State("metadata_store", "data"),
            State(PlotData.get_id(ALL, ALL), "data"),
            State(generate_id(WriteFormatDropdown), "value"),
            prevent_initial_call=True,
        )
        def download_file(
            d_clicks,
            p_clicks,
            filepath,
            df,
            aux,
            meta,
            plot_data,
            file_extension,
        ):
            df = df_from_store(df)
            aux = decode_aux(aux)

            if filepath is None or df is None:
                return dash.no_update, dmc.Notification(
                    id=str(uuid.uuid4()),
                    color="yellow",
                    title="Warning",
                    message="You have not yet loaded any data file.",
                    action="show",
                    autoClose=10000,
                )

            filepath = Path(filepath)

            with FinallyCloseBytesIO() as file:
                if ctx.triggered_id == "download-data-file-button":
                    try:
                        filename, mime = write_only_dataframe(
                            df, meta["filename"], file, file_extension
                        )
                    except Exception as err:
                        return dash.no_update, dmc.Notification(
                            id=str(uuid.uuid4()),
                            color="yellow",
                            title="Warning",
                            message=[
                                html.Div(
                                    [
                                        (
                                            "Failed to download the data file"
                                            f" for {filepath.name}"
                                        ),
                                        (
                                            html.I(" (upload)")
                                            if ctx.triggered_id
                                            == "uploaded_data_file_store"
                                            else None
                                        ),
                                        f": {err}.",
                                    ]
                                )
                            ],
                            action="show",
                            autoClose=10000,
                        )

                if ctx.triggered_id == "download-plots-file-button":
                    try:
                        for data in plot_data:
                            index = data["index"]
                            del data["index"]
                            meta["plots"][index] = data

                        filename, mime = write_dataframe_and_metadata(
                            df,
                            aux,
                            meta,
                            meta["filename"],
                            file,
                            file_extension,
                        )
                    except Exception as err:
                        return dash.no_update, dmc.Notification(
                            id=str(uuid.uuid4()),
                            color="yellow",
                            title="Warning",
                            message=[
                                html.Div(
                                    [
                                        (
                                            "Failed to download plots and data"
                                            f" file for {filepath.name}"
                                        ),
                                        (
                                            html.I(" (upload)")
                                            if ctx.triggered_id
                                            == "uploaded_data_file_store"
                                            else None
                                        ),
                                        f": {err}.",
                                    ]
                                )
                            ],
                            action="show",
                            autoClose=10000,
                        )

                encoded = base64.b64encode(file.getvalue()).decode("ascii")

                return (
                    dict(
                        base64=True,
                        content=encoded,
                        filename=filename,
                        type=mime,
                    ),
                    None,
                )

    @staticmethod
    def create_layout(data_dir=""):
        try:
            import dash_uploader as du

            uploader = du.Upload(
                id="file_uploader",
                text="Drag and Drop or Select a File to upload",
                default_style={"minHeight": 1, "lineHeight": 4},
            )
        except (ImportError, AttributeError, NotImplementedError):
            uploader = dcc.Upload(
                id="file_uploader",
                children=html.Div(
                    [
                        "Drag and Drop or ",
                        html.A("Select a File"),
                        " to upload",
                    ]
                ),
                className="dcc-upload",
            )

        return FlexRow(
            html.Div(
                [
                    layout_wrapper(
                        component=FlexRow(
                            dcc.Dropdown(
                                [
                                    {"label": fp.name, "value": str(fp)}
                                    for fp in get_data_filepaths(
                                        data_dir=data_dir
                                    )
                                ],
                                id="data_files",
                            ),
                            html.Button(
                                "Load",
                                id="submit-button",
                                n_clicks=0,
                                className="button",
                            ),
                        ),
                        title="Choose a data file",
                        css_class="dash-dropdown",
                    ),
                    html.Br(),
                    html.Div(
                        [
                            html.Div("Download the data"),
                            FlexRow(
                                html.Button(
                                    "Download only the data",
                                    id="download-data-file-button",
                                    n_clicks=0,
                                    className="button",
                                ),
                                " ",
                                html.Button(
                                    "Download plots and data",
                                    id="download-plots-file-button",
                                    n_clicks=0,
                                    className="button",
                                ),
                                WriteFormatDropdown(),
                            ),
                            dcc.Download(
                                id="data-download",
                            ),
                        ],
                    ),
                ],
                className="stretch",
            ),
            html.Div(
                [uploader],
                id="file_uploader_container",
                className="dash-uploader",
            ),
            id="control_data_content-container",
        )

    @staticmethod
    def create_layout_globals():
        return html.Div(
            [
                dcc.Store(id="uploaded_data_file_store"),
                dcc.Store(id="uploaded_auxiliary_store"),
                dcc.Store(id="uploaded_metadata_store"),
                html.Div(
                    id="data-tab-notify-container", style={"display": "none"}
                ),
                html.Div(
                    id="data-tab-upload-notify-container",
                    style={"display": "none"},
                ),
                html.Div(
                    id="data-tab-download-notify-container",
                    style={"display": "none"},
                ),
            ]
        )


class WriteFormatDropdown(dcc.Dropdown):
    def __init__(self, **kwargs):
        options = [ext for _, ext, _ in write_functions()]
        super().__init__(
            options=options,
            multi=False,
            id=generate_id(type(self)),
            placeholder="File format...",
            **kwargs,
        )


def generate_dataframe_options(upload_path, data_dir):
    return [
        {
            "label": html.Div([upload_path.name, " ", html.I("(upload)")]),
            "value": str(Path("uploads") / upload_path.name),
        }
    ] + [
        {"label": fp.name, "value": str(fp)}
        for fp in get_data_filepaths(data_dir=data_dir)
    ]



import sys
import uuid

import dash
import dash_mantine_components as dmc
import pandas as pd
from dash import Input, Output, State, ctx, dcc, html
from dash_extensions.enrich import CycleBreakerInput

from xiplot.tabs import Tab
from xiplot.utils.auxiliary import (
    CLUSTER_COLUMN_NAME,
    decode_aux,
    encode_aux,
    merge_df_aux,
)
from xiplot.utils.cluster import KMeans, cluster_colours
from xiplot.utils.components import ClusterDropdown, ColumnDropdown, FlexRow
from xiplot.utils.dataframe import get_numeric_columns
from xiplot.utils.layouts import layout_wrapper
from xiplot.utils.regex import get_columns_by_regex


class Cluster(Tab):
    @staticmethod
    def register_callbacks(app, df_from_store, df_to_store):
        @app.callback(
            Output("auxiliary_store", "data"),
            Output("cluster-tab-main-notify-container", "children"),
            Output("cluster-tab-compute-done", "children"),
            State("data_frame_store", "data"),
            State("auxiliary_store", "data"),
            Input("cluster-button", "value"),
            State("cluster-tab-compute-done", "children"),
            Input("clusters_reset-button", "n_clicks"),
            State("cluster_amount", "value"),
            State("cluster_feature", "value"),
        )
        def set_clusters(
            df,
            aux,
            process_id,
            done,
            n_clicks,
            n_clusters,
            features,
        ):
            if ctx.triggered_id == "clusters_reset-button":
                if df is None:
                    return (
                        dash.no_update,
                        dmc.Notification(
                            id=str(uuid.uuid4()),
                            color="yellow",
                            title="Warning",
                            message="You have not yet loaded any data file.",
                            action="show",
                            autoClose=10000,
                        ),
                        dash.no_update,
                    )

                df = df_from_store(df)
                notifications = []

                if process_id != done and process_id is not None:
                    notifications.append(
                        dmc.Notification(
                            id=process_id,
                            color="blue",
                            title="Info",
                            message="The clustering was cancelled.",
                            action="update",
                            autoClose=5000,
                        )
                    )

                if ctx.triggered_id == "clusters_reset-button":
                    notifications.append(
                        dmc.Notification(
                            id=str(uuid.uuid4()),
                            color="green",
                            title="Success",
                            message="The clusters were reset successfully!",
                            action="show",
                            autoClose=5000,
                        )
                    )

                aux = decode_aux(aux)
                if CLUSTER_COLUMN_NAME in aux:
                    del aux[CLUSTER_COLUMN_NAME]
                    return encode_aux(aux), notifications, process_id
                return dash.no_update, notifications, process_id

            if ctx.triggered_id == "cluster-button":
                notifications = []

                try:
                    aux = decode_aux(aux)
                    df2 = merge_df_aux(df_from_store(df), aux)
                    kmeans_col = Cluster.create_by_input(
                        df2,
                        features,
                        n_clusters,
                        notifications,
                        process_id,
                    )

                    if kmeans_col != dash.no_update:
                        aux[CLUSTER_COLUMN_NAME] = pd.Categorical(kmeans_col)
                        return encode_aux(aux), notifications, process_id
                except ImportError as err:
                    raise err
                except Exception as err:
                    notifications.append(
                        dmc.Notification(
                            id=process_id or str(uuid.uuid4()),
                            color="red",
                            title="Error",
                            message=(
                                "The clustering failed with an internal"
                                " error. Please report the following bug:"
                                f" {err}"
                            ),
                            action="update" if process_id else "show",
                            autoClose=False,
                        )
                    )
                return dash.no_update, notifications, process_id

            notifications = []

            if process_id != done and process_id is not None:
                notifications.append(
                    dmc.Notification(
                        id=process_id,
                        color="blue",
                        title="Info",
                        message="The clustering was cancelled.",
                        action="update",
                        autoClose=5000,
                    )
                )

            return dash.no_update, notifications, process_id

        @app.callback(
            Output("cluster-button", "value"),
            Output("cluster-tab-compute-notify-container", "children"),
            Input("cluster-button", "n_clicks"),
            State("cluster_amount", "value"),
            State("cluster_feature", "value"),
            State("cluster-tab-compute-done", "children"),
            State("cluster-button", "value"),
        )
        def compute_clusters(n_clicks, n_clusters, features, done, doing):
            if done != doing:
                return dash.no_update, dmc.Notification(
                    id=str(uuid.uuid4()),
                    color="yellow",
                    title="Warning",
                    message=(
                        "The k-means clustering process has not yet finished."
                    ),
                    action="show",
                    autoClose=10000,
                )

            notifications = []

            if not Cluster.validate_cluster_params(
                features, n_clusters, notifications=notifications
            ):
                return dash.no_update, notifications

            process_id = str(uuid.uuid4())

            message = "The k-means clustering process has started."

            if "sklearn" not in sys.modules:
                message += " [Loading scikit-learn]"

            return process_id, dmc.Notification(
                id=process_id,
                color="blue",
                title="Processing",
                message=message,
                action="show",
                loading=True,
                autoClose=False,
                disallowClose=True,
            )

        ColumnDropdown.register_callback(
            app,
            "cluster_feature",
            df_from_store,
            numeric=True,
            regex_button_id="add_by_keyword-button",
            regex_input_id="feature_keyword-input",
        )

        @app.callback(
            Output("selection_cluster_dropdown", "value"),
            Output("selection_cluster_dropdown", "disabled"),
            Input("cluster_selection_mode", "value"),
        )
        def pin_selection_cluster(selection_mode):
            if not selection_mode:
                return "c1", True
            return dash.no_update, False

        @app.callback(
            Output("cluster-tab-settings-session", "children"),
            Output("cluster_selection_mode", "value"),
            Output("selection_cluster_dropdown", "value"),
            Output("cluster-tab-settings-notify-container", "children"),
            CycleBreakerInput("metadata_store", "data"),
            State("cluster-tab-settings-session", "children"),
        )
        def init_from_settings(meta, last_meta_session):
            if meta is None:
                return (
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                )

            if meta["session"] == last_meta_session:
                return (
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                )

            import jsonschema

            try:
                jsonschema.validate(
                    instance=meta,
                    schema=dict(
                        type="object",
                        properties=dict(
                            settings=dict(
                                type="object",
                                properties={
                                    "cluster-tab": dict(
                                        type="object",
                                        properties=dict(
                                            selection={
                                                "type": "object",
                                                "properties": dict(
                                                    mode=dict(
                                                        enum=["fg-bg", "draw"]
                                                    ),
                                                    brush=dict(
                                                        enum=list(
                                                            cluster_colours().keys()  # noqa: 501
                                                        )
                                                    ),
                                                ),
                                                "if": dict(
                                                    properties=dict(
                                                        mode=dict(const="draw")
                                                    ),
                                                ),
                                                "then": dict(
                                                    required=["brush"]
                                                ),
                                            },
                                        ),
                                    ),
                                },
                            ),
                        ),
                        required=["settings"],
                    ),
                )
            except jsonschema.exceptions.ValidationError as err:
                return (
                    meta["session"],
                    dash.no_update,
                    dash.no_update,
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message=(
                            "Invalid cluster tab settings at"
                            f" meta{err.json_path[1:]}: {err.message}."
                        ),
                        action="show",
                        autoClose=10000,
                    ),
                )

            try:
                mode = meta["settings"]["cluster-tab"]["selection"]["mode"]
            except Exception:
                return (
                    meta["session"],
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                )

            if mode == "draw":
                brush = meta["settings"]["cluster-tab"]["selection"]["brush"]

                return meta["session"], ["edit"], brush, dash.no_update
            else:  # mode == "fg-bg"
                return meta["session"], None, dash.no_update, dash.no_update

        @app.callback(
            Output("metadata_store", "data"),
            State("metadata_store", "data"),
            Input("cluster_selection_mode", "value"),
            Input("selection_cluster_dropdown", "value"),
        )
        def update_settings(meta, selection_mode, selection_cluster):
            if meta is None:
                return dash.no_update

            if not selection_mode:
                meta["settings"]["cluster-tab"] = dict(
                    selection=dict(mode="fg-bg")
                )
            else:
                meta["settings"]["cluster-tab"] = dict(
                    selection=dict(mode="draw", brush=selection_cluster)
                )

            return meta

        ClusterDropdown.register_callbacks(
            app, "selection_cluster_dropdown", True
        )

    @staticmethod
    def validate_cluster_params(
        features, n_clusters, notifications=None, process_id=None
    ):
        if features is None or n_clusters is None:
            if notifications is not None and process_id is not None:
                notifications.append(
                    dmc.Notification(
                        id=process_id,
                        action="hide",
                    )
                )

            if features is None and notifications is not None:
                notifications.append(
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message=(
                            "You have not selected any features to cluster by."
                        ),
                        action="show",
                        autoClose=10000,
                    )
                )

            if n_clusters is None and notifications is not None:
                notifications.append(
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message=(
                            "You have not selected the number of clusters."
                        ),
                        action="show",
                        autoClose=10000,
                    )
                )

            return False

        return True

    @staticmethod
    def create_by_input(
        df,
        features,
        n_clusters,
        notifications=None,
        process_id=None,
    ):
        if not Cluster.validate_cluster_params(
            features, n_clusters, notifications, process_id
        ):
            return dash.no_update

        from sklearn.impute import SimpleImputer
        from sklearn.preprocessing import StandardScaler

        scaler = StandardScaler()
        imputer = SimpleImputer(strategy="mean")

        columns = get_numeric_columns(df)
        new_features = get_columns_by_regex(columns, features)
        x = scaler.fit_transform(df[new_features])
        x = imputer.fit_transform(x)

        km = KMeans(n_clusters=int(n_clusters)).fit_predict(x)
        kmeans_col = [f"c{c+1}" for c in km]

        notifications.append(
            dmc.Notification(
                id=process_id or str(uuid.uuid4()),
                color="green",
                title="Success",
                message="The data was clustered successfully!",
                action="update" if process_id else "show",
                autoClose=5000,
                disallowClose=False,
            )
        )

        return kmeans_col

    @staticmethod
    def create_layout():
        return html.Div(
            [
                FlexRow(
                    layout_wrapper(
                        component=FlexRow(
                            ColumnDropdown(id="cluster_feature", multi=True),
                            html.Button(
                                "Add features by regex",
                                id="add_by_keyword-button",
                                className="button",
                            ),
                        ),
                        css_class="stretch",
                        title="Features",
                    ),
                    layout_wrapper(
                        component=FlexRow(
                            dcc.Dropdown(
                                options=[
                                    i for i in range(2, len(cluster_colours()))
                                ],
                                value=3,
                                clearable=False,
                                id="cluster_amount",
                            ),
                            html.Button(
                                "Compute clusters",
                                id="cluster-button",
                                className="button",
                            ),
                        ),
                        css_class="stretch",
                        title="Number of clusters",
                    ),
                ),
                layout_wrapper(
                    component=dcc.Input(id="feature_keyword-input"),
                    style={"display": "none"},
                ),
                html.Br(),
                FlexRow(
                    dcc.Checklist(
                        [dict(label="Cluster edit mode", value="edit")],
                        id="cluster_selection_mode",
                        labelClassName="button",
                        inline=True,
                    ),
                    layout_wrapper(
                        component=ClusterDropdown(
                            id="selection_cluster_dropdown",
                            value="c1",
                            disabled=True,
                        ),
                        css_class="dash-dropdown",
                        title="Select cluster",
                    ),
                    html.Button(
                        "Remove clusters",
                        id="clusters_reset-button",
                        className="button",
                    ),
                ),
            ],
            id="control_clusters_content-container",
        )

    @staticmethod
    def create_layout_globals():
        return html.Div(
            [
                html.Div(
                    id="cluster-tab-main-notify-container",
                    style={"display": "none"},
                ),
                html.Div(
                    id="cluster-tab-regex-notify-container",
                    style={"display": "none"},
                ),
                html.Div(
                    id="cluster-tab-compute-notify-container",
                    style={"display": "none"},
                ),
                html.Div(
                    id="cluster-tab-compute-done", style={"display": "none"}
                ),
                html.Div(
                    id="cluster-tab-settings-notify-container",
                    style={"display": "none"},
                ),
                html.Div(
                    id="cluster-tab-settings-session",
                    style={"display": "none"},
                ),
            ],
            style={"display": "none"},
        )



import dash
import pandas as pd
from selenium.webdriver.common.by import By

from tests.util_test import (
    click_pdf_button,
    render_plot,
    select_dropdown,
    start_server,
)
from xiplot.plots.boxplot import Boxplot


def test_boxplot_browser(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "boxplot")

    # Check render
    plot = driver.find_element(By.CLASS_NAME, "dash-graph")
    assert "Boxplot" in plot.get_attribute("id")
    dropdowns = driver.find_element(By.ID, "plots").find_elements(
        By.CLASS_NAME, "dash-dropdown"
    )
    assert len(dropdowns) == 8

    # Change axis
    select_dropdown(dropdowns[0], "origin")
    assert "origin" in driver.find_element(By.CLASS_NAME, "xtitle").text

    # Change color
    select_dropdown(dropdowns[4], "cylinder")
    assert "cylinder" in driver.find_element(By.CLASS_NAME, "dash-graph").text

    # Change plot
    select_dropdown(dropdowns[6], "Violin")
    select_dropdown(dropdowns[6], "Strip")

    # Download pdf
    click_pdf_button(driver)

    # Close browser
    assert dash_duo.get_logs() == [], "browser console should contain no error"
    driver.close()


def test_boxplot_create():
    render = Boxplot.register_callbacks(
        dash.Dash(__name__), lambda x: x, lambda x: x
    )

    for plot in ["Box plot", "Violin plot", "Strip chart"]:
        fig = render(
            "col1",
            "col2",
            plot,
            "Clusters",
            0,
            pd.DataFrame({"col1": [1, 2], "col2": [3, 4]}),
            pd.DataFrame(index=range(2)),
            None,
        )
        assert str(type(fig)) == "<class 'plotly.graph_objs._figure.Figure'>"



import numpy as np
import pandas as pd
from dash import (
    ALL,
    MATCH,
    Input,
    Output,
    State,
    ctx,
    dash_table,
    dcc,
    html,
    no_update,
)
from dash.exceptions import PreventUpdate
from dash_extensions.enrich import CycleBreakerInput

from xiplot.plots import APlot
from xiplot.utils.auxiliary import (
    CLUSTER_COLUMN_NAME,
    SELECTED_COLUMN_NAME,
    decode_aux,
    encode_aux,
    merge_df_aux,
)
from xiplot.utils.cluster import cluster_colours
from xiplot.utils.components import (
    ColumnDropdown,
    DeleteButton,
    FlexRow,
    PlotData,
)
from xiplot.utils.regex import get_columns_by_regex
from xiplot.utils.table import get_sort_by, get_updated_item


class Table(APlot):
    @classmethod
    def register_callbacks(cls, app, df_from_store, df_to_store):
        @app.callback(
            [
                Output({"type": "table", "index": ALL}, "data"),
                Output({"type": "table", "index": ALL}, "sort_by"),
                Output({"type": "table", "index": ALL}, "selected_rows"),
            ],
            CycleBreakerInput("auxiliary_store", "data"),
            State({"type": "table", "index": ALL}, "data"),
            Input({"type": "table", "index": ALL}, "sort_by"),
            prevent_initial_call=False,
        )
        def update_table_data(aux, table_df, sort_by):
            if aux is None:
                return no_update
            aux = decode_aux(aux)

            # Try branch for testing
            try:
                trigger = ctx.triggered_id
                if trigger == "data_frame_store" or not table_df:
                    raise PreventUpdate()
            except PreventUpdate:
                raise
            except Exception:
                trigger = "auxiliary_store"

            table_data = []
            sort_bys = []

            if SELECTED_COLUMN_NAME in aux:
                selected = aux[SELECTED_COLUMN_NAME]
                where = np.where(aux[SELECTED_COLUMN_NAME])[0]
            else:
                selected = [False] * aux.shape[0]
                where = []

            for table_df, sort_by in zip(table_df, sort_by):
                table_df = pd.DataFrame(table_df)
                table_df.rename_axis("index_copy")
                table_df["Selection"] = selected

                sort_by = get_sort_by(sort_by, selected, trigger)

                table_data.append(table_df.to_dict("records"))
                sort_bys.append(sort_by)

            return table_data, sort_bys, [where] * len(sort_bys)

        @app.callback(
            Output("auxiliary_store", "data"),
            Input({"type": "table", "index": ALL}, "selected_rows"),
            State("auxiliary_store", "data"),
        )
        def update_selected_rows_store(selected_rows_checkbox, aux):
            if (
                selected_rows_checkbox == [None]
                or len(selected_rows_checkbox) == 0
            ):
                raise PreventUpdate()

            if aux is None:
                return no_update

            # Try branch for testing
            try:
                # check, which table had changed
                selected_rows_checkbox = get_updated_item(
                    selected_rows_checkbox,
                    ctx.triggered_id["index"],
                    ctx.inputs_list[0],
                )
            except Exception:
                selected_rows_checkbox = selected_rows_checkbox[0]

            aux = decode_aux(aux)
            result = [False] * aux.shape[0]
            for row in selected_rows_checkbox:
                result[row] = True
            aux[SELECTED_COLUMN_NAME] = result
            return encode_aux(aux)

        @app.callback(
            output=dict(
                cell_store=Output("lastly_clicked_point_store", "data"),
                active_cell=Output(
                    {"type": "table", "index": ALL}, "active_cell"
                ),
            ),
            inputs=[
                Input({"type": "table", "index": ALL}, "active_cell"),
                Input(
                    {"type": "table", "index": ALL}, "derived_viewport_indices"
                ),
                State("data_frame_store", "data"),
            ],
        )
        def update_lastly_activated_cell(active_cells, table_row_indices, df):
            if active_cells == [None]:
                raise PreventUpdate()

            row = None
            for cell, indices in zip(active_cells, table_row_indices):
                if cell:
                    row = indices[cell["row"]]
                    break

            return dict(cell_store=row, active_cell=[None] * len(active_cells))

        @app.callback(
            Output({"type": "table", "index": ALL}, "data"),
            Output({"type": "table", "index": ALL}, "columns"),
            Input(
                {"type": "table_columns_submit-button", "index": ALL},
                "n_clicks",
            ),
            State(cls.get_id(ALL, "columns_dropdown"), "value"),
            State({"type": "table", "index": ALL}, "columns"),
            State({"type": "table", "index": ALL}, "data"),
            State("data_frame_store", "data"),
            State("auxiliary_store", "data"),
        )
        def update_table_columns(
            n_clicks,
            dropdown_columns,
            columns,
            table_df,
            df,
            aux,
        ):
            if not ctx.triggered_id:
                raise PreventUpdate()

            trigger_id = ctx.triggered_id["index"]
            df = merge_df_aux(df_from_store(df), aux)

            for id, item in enumerate(ctx.inputs_list[0]):
                if item["id"]["index"] == trigger_id:
                    if n_clicks[id] is None or dropdown_columns[id] is None:
                        raise PreventUpdate()
                    new_columns = get_columns_by_regex(
                        df.columns.to_list(), dropdown_columns[id]
                    )
                    columns[id] = [
                        {"name": c, "id": c, "hideable": True}
                        for c in new_columns
                    ]
                    table_df[id] = df[new_columns].to_dict("records")
                    break

            return table_df, columns

        ColumnDropdown.register_callback(
            app,
            cls.get_id(MATCH, "columns_dropdown"),
            df_from_store,
            regex_button_id=cls.get_id(MATCH, "regex_button"),
            regex_input_id=cls.get_id(MATCH, "regex_input"),
        )

        PlotData.register_callback(
            cls.name(),
            app,
            dict(
                page=Input({"type": "table", "index": ALL}, "page_current"),
                query=Input({"type": "table", "index": ALL}, "filter_query"),
                sort_by=Input({"type": "table", "index": ALL}, "sort_by"),
                hidden=Input(
                    {"type": "table", "index": ALL}, "hidden_columns"
                ),
                columns=Input({"type": "table", "index": ALL}, "columns"),
                dropdown=Input(cls.get_id(MATCH, "columns_dropdown"), "value"),
            ),
            lambda inputs: dict(
                columns={
                    column: {
                        "hidden": column in inputs["hidden"],
                        **{
                            "sorting": sb["direction"]
                            for sb in inputs["sort_by"]
                            if column == sb["column_id"]
                        },
                    }
                    for column in [c["id"] for c in inputs["columns"]]
                },
                query=inputs["query"] or "",
                page=inputs["page"] or 0,
                dropdown=inputs["dropdown"],
            ),
        )

        return [
            update_table_data,
            update_selected_rows_store,
            update_lastly_activated_cell,
            update_table_columns,
        ]

    @classmethod
    def create_new_layout(cls, index, df, columns, config=dict()):
        import jsonschema

        df = df.rename_axis("index_copy")
        columns = df.columns.to_list()
        jsonschema.validate(
            instance=config,
            schema=dict(
                type="object",
                properties=dict(
                    columns=dict(
                        type="object",
                        additionalProperties=dict(
                            type="object",
                            properties=dict(
                                hidden=dict(type="boolean"),
                                sorting=dict(enum=["asc", "desc"]),
                            ),
                        ),
                    ),
                    query=dict(type="string"),
                    page=dict(type="integer", minimum=0),
                ),
            ),
        )

        if "columns" in config:
            columns = list(config["columns"].keys())
            hidden_columns = [
                c
                for c, v in config["columns"].items()
                if v.get("hidden", False)
            ]
            sort_by = [
                dict(column_id=c, direction=v["sorting"])
                for c, v in config["columns"].items()
                if "sorting" in v
            ]
        else:
            columns = columns[:5]
            hidden_columns = []
            sort_by = []

        filter_query = config.get("query", "")
        page_current = config.get("page", 0)

        for c in columns:
            if type(df[c][0]) == np.ndarray:
                df = df.astype({c: str})
        return html.Div(
            children=[
                DeleteButton(index),
                dash_table.DataTable(
                    id={"type": "table", "index": index},
                    columns=[
                        {"name": c, "id": c, "hideable": True} for c in columns
                    ],
                    hidden_columns=hidden_columns,
                    data=df[columns].to_dict("records"),
                    editable=False,
                    page_action="native",
                    page_size=25,
                    page_current=page_current,
                    sort_action="native",
                    sort_mode="multi",
                    sort_by=sort_by,
                    row_selectable="multi",
                    filter_action="native",
                    filter_query=filter_query,
                    fixed_rows={"headers": True},
                    style_as_list_view=True,
                    style_table={"overflowX": "auto"},
                    style_cell={
                        "height": "auto",
                        "minWidth": "180px",
                        "width": "180px",
                        "maxWidth": "180px",
                        "whiteSpace": "normal",
                        "wordBreak": "break-all",
                    },
                    style_data_conditional=[
                        {
                            "if": {
                                "filter_query": (
                                    f'{{{CLUSTER_COLUMN_NAME}}} = "{cluster}"'
                                ),
                            },
                            "backgroundColor": colour,
                            "color": "black",
                        }
                        for cluster, colour in cluster_colours().items()
                        if cluster != "all"
                    ],
                ),
                FlexRow(
                    ColumnDropdown(
                        cls.get_id(index, "columns_dropdown"),
                        multi=True,
                        options=df.columns,
                        placeholder="Select Columns...",
                        value=config.get("dropdown", None),
                    ),
                    html.Button(
                        "Add columns by regex",
                        id=cls.get_id(index, "regex_button"),
                        className="button",
                    ),
                    html.Button(
                        "Update table",
                        id={
                            "type": "table_columns_submit-button",
                            "index": index,
                        },
                        className="button",
                    ),
                ),
                dcc.Input(
                    id=cls.get_id(index, "regex_input"),
                    style={"display": "none"},
                ),
                PlotData(index, cls.name()),
            ],
            id={"type": "table-container", "index": index},
            className="plots",
        )



class ServerSideStoreBackend:
    def __init__(self):
        self.store = dict()

    def get(self, key, ignore_expired=False):
        return self.store.get(key)

    def set(self, key, value):
        self.store[key] = value

    def has(self, key):
        return key in self.store



from abc import ABC, abstractstaticmethod


class Tab(ABC):
    def __init__(self):
        raise TypeError("Tabs should not be constructed")

    @classmethod
    def name(cls) -> str:
        return cls.__name__

    @abstractstaticmethod
    def register_callbacks(app, df_from_store, df_to_store):
        pass

    @abstractstaticmethod
    def create_layout():
        pass

    @abstractstaticmethod
    def create_layout_globals():
        pass



import dash
import pandas as pd
from selenium.webdriver.common.by import By

from tests.util_test import render_plot, start_server
from xiplot.plots.heatmap import Heatmap

render = Heatmap.register_callbacks(
    dash.Dash(__name__), lambda x: x, lambda x: x
)[0]


def test_tehe001_render_heatmap(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Heatmap")

    plot = driver.find_element(By.CLASS_NAME, "dash-graph")

    assert "heatmap" in plot.get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_create_heatmap():
    d = {"col1": [1, 2], "col2": [3, 4]}
    df = pd.DataFrame(data=d)
    output = render(2, ["col1", "col2"], 1, df, pd.DataFrame())
    fig = output

    assert str(type(fig)) == "<class 'plotly.graph_objs._figure.Figure'>"



import time

from xiplot.setup import setup_xiplot_dash_app


def test_dash001_launch(dash_duo):
    dash_duo.start_server(setup_xiplot_dash_app(data_dir="data"))
    time.sleep(1)
    dash_duo.wait_for_page()

    assert dash_duo.get_logs() == [], "browser console should contain no error"



from typing import Optional, Sequence, Union

import pandas as pd

CLUSTER_COLUMN_NAME = "Xiplot_cluster"
SELECTED_COLUMN_NAME = "Xiplot_selected"


def get_clusters(aux: pd.DataFrame, n: Optional[int] = None) -> pd.Categorical:
    """Get the cluster column from the auxiliary data.

    Args:
        aux: Auxiliary data frame.
        n: Column size if missing. Defaults to `aux.shape[0]`.

    Returns:
        Categorical column with clusters (creates a column with `n` "all" if missing)
    """
    if not isinstance(aux, pd.DataFrame):
        aux = decode_aux(aux)
    if CLUSTER_COLUMN_NAME in aux:
        return pd.Categorical(aux[CLUSTER_COLUMN_NAME].copy())
    if n is None:
        n = aux.shape[0]
    return pd.Categorical(["all"]).repeat(n)


def get_selected(aux: pd.DataFrame, n: Optional[int] = None) -> pd.Series:
    """Get the selected column from the auxiliary data.

    Args:
        aux: Auxiliary data frame.
        n: Column size if missing. Defaults to `aux.shape[0]`.

    Returns:
        Column with booleans (creates a column with `[False] * n` if missing)
    """
    if not isinstance(aux, pd.DataFrame):
        aux = decode_aux(aux)
    if SELECTED_COLUMN_NAME in aux:
        return aux[SELECTED_COLUMN_NAME].copy()
    if n is None:
        n = aux.shape[0]
    return pd.Series([False]).repeat(n).reset_index(drop=True)


def toggle_selected(
    aux: pd.DataFrame, rows: Sequence[int], n: Optional[int] = None
) -> pd.DataFrame:
    """Toggle rows in the selected column in the auxiliary data.

    Args:
        aux: Auxiliary data frame.
        rows: Rows to toggle.
        n: Column size if missing. Defaults to `aux.shape[0]`.

    Returns:
        Updated auxiliary data frame.
    """
    encode = not isinstance(aux, pd.DataFrame)
    if encode:
        aux = decode_aux(aux)
    selected = get_selected(aux, n)
    if isinstance(rows, int):
        rows = (rows,)
    for row in rows:
        selected[row] = not selected[row]
    aux[SELECTED_COLUMN_NAME] = selected
    if encode:
        aux = encode_aux(aux)
    return aux


def decode_aux(aux: str) -> pd.DataFrame:
    if isinstance(aux, pd.DataFrame):
        return aux
    return pd.read_json(aux, orient="table")


def encode_aux(aux: pd.DataFrame) -> str:
    return aux.to_json(orient="table", index=False)


def merge_df_aux(
    df: pd.DataFrame, aux: Union[str, pd.DataFrame]
) -> pd.DataFrame:
    if not isinstance(aux, pd.DataFrame):
        aux = decode_aux(aux)
    aux.index = df.index
    return pd.concat((df, aux), axis=1)
